!=======================================================================
!                           COMPUTE_NEIKDTREE
!=======================================================================
! Author : S. Colombi
!          Institut d'Astrophysique de Paris
!          98 bis bd Arago, F-75014, Paris, France 
!          colombi@iap.fr
!
! This program has multiple usage and can do three things
! (1) Read an input particle distribution file and compute the mean
!     square distance between each particle and its nearest neighbourgs
! (2) Read an input particle distribution file and compute the SPH
!     density associated to each particle + the list of its nearest
!     neighbourgs
! (3) Read an input particle distribution and a neighbourgs file which
!     is an output of step 2), and output the tree of the structures in
!     structures.
!
! If steps (1) and (2) are probably correct, step (3) is still under
! test and construction. Step (1) was rather extensively tested, while
! step (2) needs further tests although I am quite confident it should
! be correct.
!
!=======================================================================
!
! USAGE : 
! +++++
!    
! compute_neikdtree config_file_name
! or
! compute_neikdtree 
! In the last case the default config file name is used, i.e.
! compute_neiKDtree.config
! The config file contains all the informations required for
! running compute_neikdtree
! The FORTRAN namelist technique is used, where the parameters are
! specified by ``name=parameter''
!
!=======================================================================
!
! PARAMETERS IN THE CONFIG FILE 
! +++++++++++++++++++++++++++++
! 
! A example of config file is given in compute_neiKDtree.config
!
!
! verbose     : normal verbose mode (.true. or .false.)
! megaverbose : full verbose mode (.true. or .false.)
! filein      : name of the input particles/velocities file ('myfilename')
! Ntype       : format of the file (integer number)
!               0 : PM simple precision unformatted
!               1 : GADGET simple precision unformatted, cosmological simulation
!                   of dark matter only with periodic boundaries
!               2 : RAMSES unformatted (particles) simulation
!                   MPI multiple output
!               3 : CONE unformatted 
!               4 : Ninin treecode unformatted
! remove_degenerate : if this set to .true., check for particles at the same 
!               position and add a random displacement at approximately 
!               floating representation accuracy in order to avoid infinite
!               number of KD tree cells creation. This parameter is global since
!               if a random displacement has been applied once for computing
!               SPH density, the same one must be applied again for subsequent 
!               calculations in order to be self-consistent.
!               Setting remove_degenerate=.false. will of course speep-up
!               the program at the cost of a risk of crash due to infinite
!               KD tree cell creation. This is however safe in most cases since
!               KD tree creation can now deal with particles at the same position.
!               It is therefore recommended to set remove_degenerate to .false.
!               first before trying .true.
! action      : 'distances' for computing mean square distance between each
!                           particle and its nearest neighbourgs
!               'neighbors' for computing the list of the nearest neighbourgs of
!                           each particle and its SPH density
!               'adaptahop' for computing the tree of structures and substructures
!                           in the simulation 
!               'posttreat' for computing the physical properties of dynamically
!                           selected haloes and subhaloes
!
! if (Ntype=1) : GADGET format
! --------------
!
! nmpigadget  : number of CPU used to perform the simulation (in order to read the 
!               appropriate number of files). If there is no multiple file 
!               extension (only one file), set nmpigadget=-1.
!
! if (Ntype=2) : RAMSES unformatted (dark matter particles)
! --------------
!
! mtot        : total mass in the full simulation box, of the population considered,
!               in internal RAMSES units.
!               mtot=-1 works if full simulation box is analysed.
! Ltot        : size of the full simulation box, in internal RAMSES units.
!               Ltot=-1 gives default value, valid for cosmological simulations
!               only.
! 
! if (Ntype=3) : CONE unformatted (dark matter particles) 
! ------------------
!
! boxsize     : comoving size of the simulation box in Mpc
!
! if (Ntype=4) : Ninin treecode unformatted 
! --------------
!
! boxsize2    : comoving size of the simulation box, in Mpc
! hubble      : value of H0/100 in km/s/Mpc
! omega0      : value of cosmological density parameter
! omegaL      : value of cosmological constant
! aexp_max    : value of expansion factor at present time
!
! if (action='distances') :
! -------------------------
!
! nvoisdis    : number of neighbors considered for each particle
! filedis     : name of the output file for the mean square distances 
!               ('mydistancename')
!               (this is a binary unformatted or formatted file which will be 
!               described more in detail in the near future) 
! formatted   : .true. if the output file is in ASCII, .false. if the output
!               file is in binary unformatted double float.
! ncpu        : number of virtual threads for a parallel calculation with openMP
!               This integer number should be a multiple of the real number of 
!               threads used during the run. The larger it will be, the better
!               the load balancing will be. 
! velocities  : .true. or .false. : if set to .true., include velocities 
!               treatment in the calculation and the output of the results
!
! if (action='neighbors') :
! -------------------------
!
! nvoisnei    : number of neighbors considered for computing the SPH density
!               (integer number). Typically this number should vary between
!               20 and 100 (template value 64).
! nhop        : number of stored nearest neighbourgs in the output file
!              (integer number).
!              nhop must smaller smaller or equal to nvoisnei. Typically nhop
!              should be of order 10-30 (template value 16).
! fileneinei  : name of the output file for the SPH density and the list of
!              nhop nearest neighbors for each particle ('myfileneighbors')
!              (this is a binary unformatted file which will be described more 
!              in detail in the near future)
!
! if (action='adaptahop') :
! -------------------------
!
! rho_threshold : density threshold. Particles with SPH density below this
!              density threshold are not selected. This thresholding will
!              define a number of connected regions, each of which corresponding
!              to a structure. For each of these structures, we aim to 
!              build a tree of substructures. Typically, one is interested
!              in finding the substructures of highly nonlinear objects. A value
!              often taken for rho_threshold is 80, which corresponds roughly
!              to friend-of-friend algorithm parameter b=2.
! nmembthresh: threshold on the number of particles that a structure or a
!              a substructure (above some density threshold rhot) must contain 
!              for being considered as significant.
!              The choice of nmembthresh is related to effects of N-body relaxation
!              (treecode) or over-softening of the forces (AMR or PM codes) or
!              SPH smoothing (SPH smoothing over a number N of particles tends
!              to ``erase'' structures with less than N particles).
!              Basically, due to either of these effects, structures or 
!              substructures with a number of particles smaller than nmembthresh
!              are not considered. A typical choice of nmembthresh is 64.
! fudgepsilon: This parameter can be seen as an Eulerian version of the thresholding
!              controlled by nmembthresh. It defines the size of the smallest structures
!              that can exist physically and is related to force softening.
!              Basically, if epsilon is the softening parameter, substructures with
!              typical radius smaller than epsilon are just coincidences and should
!              not be considered. The typical radius of a structure is given by the
!              mean square distance of particles belonging to it with respect to its
!              center of gravity. 
!              The criterion for selecting a structure is thus
!              
!              radius_{substructure} > epsilon,
!             
!              where epsilon=fudgepsilon*L/Npart^{1/3}. Fudgepsilon is thus expressed
!              in units of mean interparticle separation.
!              - For a treecode, fudgepsilon is typically of the order of 1/20;
!              - For a PM code, fudgepsilon is typically of the order of N_g/Npart^{1/3}
!              where N_g is the resolution of the grid used for the force calculations,
!              i.e. of the order of 1 or 0.5.
!              - For a quasi-Lagrangian code such as RAMSES, putting constrains from fudgespilon 
!              is not really useful. Basically it corresponds to N_g/Npart^{1/3} where N_g
!              would be the equivalent of the PM grid probed by the smallest AMR cells.              
! alpha      : a criterion can be applied to decide weither a substructure
!              is selected or not. Here the choice of alpha dictates by how
!              much the maximum local density in this substructure should be larger
!              than the local average density in this substructure:
!
!              rho_max_{substructure} >= alpha*< rho >_{substructure}.
!
!              Basically, the choice of alpha dictates the ``peakyness'' of a 
!              substructure. For instance, a substructure might contain itself
!              5 local maxima. For this substructure to be dynamically significant
!              we want the largest of the local maxima to be at least alpha
!              times the mean local density, i.e. the average density of all the
!              particles within the substructure. Dynamically bounded substructures
!              are expected to be very peaky. A typical choice of alpha would be
!              alpha=0 or a few unites, e.g. alpha=4.
! fudge      : a criterion can be applied to decide wither a substructure is 
!              statistically significant in terms of Poisson noise. Indeed,
!              even if Poisson noise is considerably reduced by SPH smoothing
!              it is still present to some extent, and some substructures might
!              be simply due to local Poisson fluctuations. 
!              Here the choice of fudge dictates by how many ``sigma's'' 
!              a structure must be compared to a random Poisson fluctuation.
!              The criterion applied is
!
!              < rho >_{substructure} > rhot*[1+fudge/sqrt(N)]
!
!              where N is the number of particles in the substructure and rhot
!              the density threshold corresponding to this substructure (in 
!              other worlds, rhot is the minimum density of particles whithin
!              the substructure). 
!              This criterion can be understood as follows : if a given substructure
!              contains N particles, the corresponding random poisson fluctuations are
!              of the order of sqrt(N). So typically the uncertainty on the density
!              estimate of a structure containing N particles is of the order of
!              sqrt(N) (this of course neglects the fact that SPH softening reduces
!              considerably the effects of poisson fluctuations). For the
!              fluctuation associated to the substructure to be significant, we
!              impose that (< rho >_{substructure} - rhot)/rhot > fudge*sigma with
!              sigma=1/sqrt(N)=sqrt(< (M-<M>)^2 >)/<M> where M is a Poisson process
!              of average N.
!              A typical value is fudge=4.
!
!              IMPORTANT NOTE : we should always have fudge > 0. Small values of
!              fudge will slow down the program, while large values of fudge will
!              make the hierarchical decomposition in terms of substructures less
!              accurate. 
!              The reason for that is that despite the fact we know all the saddle points
!              connecting all the substructures, it was impossible to figure out a simple
!              way of sorting these saddle points in order to construct automatically
!              the tree of structures and substructures. As a result, we operate 
!              iteratively by increasing the local threshold density as follows:
!              [rhot NEXT] = [rhot OLD]*[1+fudge/sqrt(N)] where N is the number of particles
!              in the substructure. Then we see weither saddle points within this
!              substructure are below the new value of rhot: if this happens, it
!              means that at the density level [rhot NEXT], the substructure is composed
!              of disconnected subsubstructures and the threshold value of connection
!              between these subsubstructures is approximated by [rhot NEXT] (the real value
!              should be between [rhot NEXT] and [rhot OLD]).
! filenode   : output file for the tree of structures and substructures. See below
!              for explanation of its format
! simu_unitsnei : .true. or .false : to specifie if the nodes positions and radii in the
!              file filenode are specified in the same units as in the simulation input
!              file or in Mpc
! filepartnodenei : ouput file for particule node number. To each particle, a
!              integer is associated, which is the node number the deepest possible in the tree
!              given in filenode. This way, at any level in the tree, it will be possible to
!              find recursively all the particles belonging to the tree. If the node number 
!              is zero, it means the SPH density of particle is below the threshold rho_threshold 
!              and is therefore associated to no structure. 
! formattedpartnodenei : .true. for having filepartnodenei in ascii format, .false. for
!              having filepartnodenei in binary format. The format is in both
!              case, the number of particles npart in the first line, followed by id(:), 
!              where id(:) is an array of npart integers.
! fileneihop : input file with SPH density and list of nearest neighbors of each
!              particle, obtained by running comptu_neiKDtree with action='neighbors'.
!
! if (action='posttreat') :
! -------------------------
! This option is still in development and test phase
!
!=======================================================================
!
! FORMAT OF OUTPUT FILES
! ++++++++++++++++++++++
!
! filenode :
! ----------
! nnodes,rho_threshold,nmembthresh,alpha,fudge,fudgepsilon,if_simu_units
! 1, node(id=1)
! [space]
! 2, node(id=2)
! [space]
! ...
! nnodes, node(id=nnodes)
! [space]
!
! where node(id) is a structure defined as follow (in order) :
!
! level : integer number corresponding to the level in the tree.
!         The root, node(1), corresponds to level=0. The root corresponds
!         to all the selected particles (above the density threshold) 
!         in the simulation box: it is not physically relevant.
! mother : the id of the mother of this node, i.e. the id of substructure
!         which contains this substructure (for the root, there is no
!         mother, i.e. mother=0).
! firstchild : the id of the first child of this node, i.e. the first
!         substructure it contains. Other substructures will be found
!         by running over the list their sisters. If we are sitting
!         at the end of a branch, firstchild=0 (i.e. this node correspond
!         to an elementary substructure) .
! nsisters : the number of sisters of this substructure + 1. 
! sister : the id of the sister ``at the right'' of this structure. A structure
!         that has not sister at its ``right'' has sister=0.
! rho_saddle : the density of the saddle(s) point(s) connecting this  
!         structure to its sisters. In principle this could be calculated
!         uniquely so that a structure would always have at most one
!         sister, but as explained in the important note above, 
!         rho_saddle is computed approximately, as a result, it can
!         correspond to more than a single saddle point.
!         For a given density threshold, rho_th, if rho_th > rho_saddle,
!         the substructure is connected to its sisters, whereas if
!         rho_th < rho_saddle, the substructure is disconnected from its
!         sisters.
! density : average density of the substructure, calculated from the
!         average SPH density of all particles with SPH density larger than
!         rho_saddle and belonging to the substructure.
! densmax : maximum density within the substructure : it correspond 
!         to the largest SPH density of all the particles within the 
!         substructure.
! radius : size in Mpc or in simulation units of the substructure calculated 
!         by computing the mean
!         square distance of each particle belonging to it with respect
!         to its center of mass.
! mass :  number of particles within the substructure.
! truemass : normalized mass within the substructure (same as mass if all
!         the particles have the same mass).
! position(3) : coordinates in Mpc of the center of mass of the 
!         substructure.
!
!
! fileneinei (as well as input file fileneihop)
! ----------
! This a fortran binary unformatted file with:
! npart,nvoisins,nhop
! (density(ipar),ipar=1,npart)
! do ipar=1,npart
!    (iparneigh(ihop,ipar),ihop=1,nhop)
! enddo
!
! npart    : number of particles in the data file (int*4)
! nvoisins : number of neighbors used to perform the SPH smoothing (int*4)
! nhop     : number of neighbors actually kept for each particle (int*4)
! density  : SPH density for each particle (assuming that average density
!           of the universe is unity) (real*4)
! iparneigh(1:nhop,ipar) : the list of nhop nearest neighbors for particle ipar
!           (int*4)
!
!
! filepartnodenei 
! ---------------
! This is either a fortran binary unformatted file or an ascii file
! according to the choice of the user.
! npart
! (partnode(ipar),ipar=1,npart)
!
! npart : number of particles in the data file (int*4)
! partnode(ipar) : the leaf to which this particle belongs (or the
!         node number the deepest possible in the tree of structures
!         and substructures) (int*4)
!         partnode(ipar)=0 when this particle belongs to nothing.
!
!=======================================================================
!
! TO COMPILE THE PROGRAM
! ++++++++++++++++++++++
!
! on a DEC ALPHA the appropriate command is
! f90 -O -recursive -o compute_neiKDtree compute_neiKDtree_common.f90 \
!                      compute_neiKDtree.f90
!
!=======================================================================
!
! HISTORY
! +++++++
!
! 02/05/02/SC/IAP : first operational clean version with namelist and
!                   three options for the calculations.
!                   + action='distances' tested by comparison with the
!                   results of a different algorithm based on link list
!                   approach instead of KDtree (compute_neighbourgs2.f)
!                   The agreement between the 2 softwares is good within
!                   float calculation accuracy (5.10^{-6} relative 
!                   maximum differences). The differences can be 
!                   explained by a slightly different approach used
!                   in the 2 programs to do floating calculations.
!                   + action='neighbors' is not fully tested but should be
!                   okay. 
!                   + action='adaptahop' has been seen to work fine on
!                   a few data sets, but is not by any mean extensively
!                   tested. (No comments on the parameters of the namelist
!                   to avoid the users using this option)
!                   --> Version 0.0
! 04/05/02/SC/IAP : Full comments on the namelist parameters corresponding
!                   to action='adaptahop'. A few comments are added to
!                   the program. A new input particle file format is 
!                   added (RAMSES). 
!                   --> Version 0.1
! 04/17/02/SC&RT/IAP : Clean up the RAMSES format. Speed up the SPH 
!                   smoothing procedure
!                   --> Version 0.2
! 12/02/02/SC/IAP : Parallelize action='distances' and action='neighbors'
!                   Add RAMSES dark matter format with periodic boundaries
!                   (Ntype=-2), add CONE format (Ntype=3)
!                   --> Version 0.3
! 27/05/03/SC&RT/IAP&STR : Add RAMSES MPI dark matter + stars format with
!                   periodic boundaries (still under work, Ntype=-20) and
!                   a subroutine to remove degeneracy in particle positions
!                   Add GADGET multiple file MPI format.
! 04/06/03/SC/IAP : Dynamical selection of haloes and subhaloes :
!                   + new action='posttreat'
! 26/10/03/SC/IAP : add Ninin treecode format Ntype=4
! 16/10/06/SC/IAP : Change the treecode algorithm to be able to deal with
!                   2 or more particles at the same position. It is not
!                   yet completely safe in terms of the calculation of
!                   the SPH density (if all the particles are at the
!                   same position one gets 0/0).
!                   Add a few comments on file formats.
!                   --> Version 0.8
! 18/10/06/SC&RT/IAP : remove obsolete formats and improve RAMSES MPI
!                   format (now swapped from -20 to 2).
!
!=======================================================================
!
! ERRORS
! ++++++
!
! If one of the following errors occures :
!
! 'ERROR in create_nodes :'
! 'All subgroups are below the threshold.'
!
! You can comment the corresponding STOP and cross your fingers.
! I did not figure out exactly what happens in that case, nothing really
! bad, apparently :0).
!
!=======================================================================
program compute_neiKDtree
!=======================================================================
  use compute_neiKDtree_common
  implicit none

  call default_parms
  call input_parms
  if (verbose) call list_parameters
  call read_part
  if (remove_degenerate) call remove_degenerate_particles
  if (outdis) then
     call create_tree_structure
     call compute_dist
  elseif (outnei) then
     call create_tree_structure
     call compute_mean_density_and_np
     call out_nei_info
  elseif (outgtree) then
     call read_nei_info
     call find_local_maxima
     call create_group_tree
     call out_group_tree
     if (partinfo) call out_partnode
  elseif (posttreatment) then
     call in_partnode
     call in_group_tree
     call posttreat
     call out_group_tree_dyn
     if (partinfo) call out_partnode
  endif
  if (verbose) write(*,*) 'END of compute_neiKDtree'
end program compute_neiKDtree

!=======================================================================
subroutine default_parms
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  config='compute_neiKDtree.config'
  verbose=.true.
  megaverbose=.false.
  periodic=.true.
  boxsize=1.
  velocities=.false.
  remove_degenerate=.false.
  action='titi'
  outdis=.false.
  outnei=.false.
  outgtree=.false.
  posttreatment=.false.
  partinfo=.false.
  Ntype=0
  filein='titi'
  filenei='tata'
  filenode='toto'
  filepartnode='tutu'
  fileinpartnode='tyty'
  filenodedyn='tonton'
  filedis='tete'
  simu_units=.true.
  formatted=.true.
  formattedpartnode=.true.
  formattedinpartnode=.true.
  nvoisins=64
  nhop=16
  rho_threshold=80.+1.
  nmembthresh=64
  alpha=10.
  fudge=4.
  fudgepsilon=0.
  ncpu=8
  nmpi=8
  niterations=1
  boxsize2=150.
  hubble=0.6666667
  aexp_max=36.587
  omega0=0.66667
  omegaL=0.33333
end subroutine default_parms

!=======================================================================
subroutine input_parms
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: nvoisdis,nhopnei,nvoisnei,nhophop,nvoishop
  integer :: nmpigadget
  character*255 :: fileneinei,fileneihop,filenodenei
  character*255 :: filepartnodenei,filepartnodedyn
  logical :: simu_unitsnei,simu_unitspos,formattedpartnodenei
  integer :: iargc
  character*255 argument

  namelist/verbosity/verbose,megaverbose
  namelist/inputfile/Ntype,filein,remove_degenerate
  namelist/ifNtype1/nmpigadget
  namelist/ifNtype2/mtot,Ltot
  namelist/ifNtype3/boxsize
  namelist/ifNtype4/boxsize2,hubble,aexp_max,omega0,omegaL
  namelist/actionp/action
  namelist/ifdistances/nvoisdis,filedis,formatted,ncpu,velocities
  namelist/ifneighbors/nhopnei,nvoisnei,fileneinei
  namelist/ifadaptahop/rho_threshold,nmembthresh,alpha,fudge,fudgepsilon,filenode, &
 &                     fileneihop,filepartnodenei,simu_unitsnei,formattedpartnodenei
  namelist/ifposttreat/fileinpartnode,filepartnodedyn,filenodedyn,filenodenei, &
 &                     simu_unitspos,niterations,formattedpartnode,formattedinpartnode



  if (iargc() == 1) then
     call getarg(1,config)
  elseif (iargc() > 1) then
     call error_usage
  endif

  open(unit=lsin,file=config,form='formatted',status='old',err=9)

  rewind(lsin)
  read(lsin,verbosity,end=1)

1 rewind(lsin)
  read(lsin,inputfile,end=20)

20 rewind(lsin)
  nmpigadget=nmpi
  read(lsin,ifNtype1,end=21)
21 if (Ntype==1) nmpi=nmpigadget
  
  rewind(lsin)
  read(lsin,ifNtype2,end=22)

22 rewind(lsin)
  read(lsin,ifNtype3,end=2)

2 rewind(lsin)
  read(lsin,ifNtype4,end=24)

24 rewind(lsin)
  read(lsin,actionp,end=3)

3 rewind(lsin)
  if (action == distances) then
     nvoisdis=nvoisins
     read(lsin,ifdistances,end=4)
4    nvoisins=nvoisdis
     outdis=.true.
  elseif (action == neighbors) then
     nvoisnei=nvoisins
     nhopnei=nhop
     fileneinei=filenei
     read(lsin,ifneighbors,end=5)
5    nvoisins=nvoisnei
     nhop=nhopnei
     filenei=fileneinei
     outnei=.true.
  elseif (action == adaptahop) then
     fileneihop=filenei
     filepartnodenei=filepartnode
     simu_unitsnei=simu_units
     formattedpartnodenei=formattedpartnode
     read(lsin,ifadaptahop,end=6)
6    filenei=fileneihop  
     filepartnode=filepartnodenei
     formattedpartnode=formattedpartnodenei
     if (filepartnode(1:1)/='#') partinfo=.true.
     outgtree=.true.
     simu_units=simu_unitsnei
     if (fudge <= 0.d0) then
        write(*,*) 'ERROR in input_parms :'
        write(*,*) 'We should have fudge > 0'
        write(*,*) 'fudge =',fudge
        STOP
     elseif (fudge <= 1.d0.and.verbose) then
        write(*,*) 'WARNING in input_parms :'
        write(*,*) 'You took a small value of parameter fudge'
        write(*,*) 'This might slow down considerably the calculation'
     endif
  elseif (action == posttreatm) then
     write(*,*) 'Action = '//posttreatm//' not yet available.'
     STOP
     posttreatment=.true.
     velocities=.true.
     simu_unitspos=simu_units
     filepartnodedyn=filepartnode
     filenodenei=filenode
     read(lsin,ifposttreat,end=7)
 7   filepartnode=filepartnodedyn
     filenode=filenodenei
     if (filepartnode(1:1)/='#') partinfo=.true.
     simu_units=simu_unitspos
  else
     write(*,*) 'Found nothing to do...'
     STOP
  endif

  close(lsin)
  return

9 write(*,*) 'ERROR in input_parms :'
  write(*,*) 'Impossible to open the following config file :'
  write(*,*) trim(config)
  STOP

end subroutine input_parms

!=======================================================================
subroutine error_usage 
!=======================================================================
  implicit none 
  write(*,*) 'USAGE : compute_neiKDtree config_file_name'
  STOP
end subroutine error_usage

!=======================================================================
subroutine list_parameters
!=======================================================================
  use compute_neiKDtree_common
  implicit none

  write(*,*) '============================================================'
  write(*,*) 'Compute_neiKDtree was called with the following attributes :'
  write(*,'(a)')    ' config file   : '//trim(config)
  write(*,*)        'verbose       :',verbose
  write(*,*)        'megaverbose   :',megaverbose
  write(*,'(a)')   ' filein        : '//trim(filein)
  write(*,*)        'Ntype         :',Ntype
  write(*,*)        'remove_degen. :',remove_degenerate
  if (Ntype==1)write(*,*)     'nmpi          :',nmpi
  if (Ntype==2) then
     write(*,*)     'mtot          :',mtot
     write(*,*)     'Ltot          :',Ltot
  endif
  if (Ntype.eq.3) then
     write(*,*)     'boxsize       :',boxsize
  endif
  if (Ntype.eq.4) then
     write(*,*)     'boxsize2      :',boxsize2
     write(*,*)     'hubble        :',hubble
     write(*,*)     'omega0        :',omega0
     write(*,*)     'omegaL        :',omegaL
     write(*,*)     'aexp_max      :',aexp_max
  endif
  if (outdis) then
     write(*,'(a)')' Action        : '//distances
     write(*,*)     'nvoisins      :',nvoisins
     write(*,*)     'velocities    :',velocities
     write(*,'(a)')' filedis       : '//trim(filedis)
     write(*,*)     'formatted     :',formatted
  elseif (outnei) then
     write(*,'(a)')' Action        : '//neighbors
     write(*,*)     'nvoisins      :',nvoisins
     write(*,*)     'nhop          :',nhop
     write(*,'(a)')' filenei       : '//trim(filenei)
  elseif (outgtree) then
     write(*,'(a)')' Action        : '//adaptahop
     write(*,*)     'rho_threshold :',rho_threshold
     write(*,*)     'nmembthresh   :',nmembthresh
     write(*,*)     'alpha         :',alpha
     write(*,*)     'fudge         :',fudge
     write(*,*)     'fudgepsilon   :',fudgepsilon
     write(*,*)     'simu_units    :',simu_units
     write(*,'(a)')' filenode      : '//trim(filenode)
     write(*,'(a)')' filepartnode  : '//trim(filepartnode)
     write(*,'(a)')' filenei       : '//trim(filenei)
  elseif (posttreatment) then
     write(*,'(a)')' Action        : '//posttreatm
     write(*,*)     'simu_units    :',simu_units
     write(*,*)     'Formattedpartn:',formattedpartnode
     write(*,'(a)')' filenodedyn   : '//trim(filenodedyn)
     write(*,'(a)')' fileinpartnode: '//trim(fileinpartnode)
     write(*,'(a)')' filepartnoded.: '//trim(filepartnode)
  endif
  write(*,*) '============================================================'
end subroutine list_parameters

!=======================================================================
subroutine posttreat
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inode,incpart,ipar,nsubpart,npartleftover,ids,ispar,inodesub
  integer :: inodefound,inodes
  real(kind=8) :: xpn(3),pad(3),diffpos(3),bindingener
  logical :: something_wrong,isbound,stopit
  logical, allocatable, dimension(:) :: touched
  integer, allocatable, dimension(:) :: idspart,idspartnode,idspartleve,idsortnode
  integer :: iteration,incchange,Niterationseff,inceliminated
  real(kind=8) :: HsurH0

  if (verbose) write(*,*) 'Dynamical posttreatment of the haloes and their substructures'

  ! Initialize various physical constants
  GMphys=mass_in_kg*gravitational_constant
  aexp_hubble_const=aexp*HsurH0(1.d0/aexp-1.d0,omega0,omegaL,1.d0-omega0-omegaL)*hubble*100.d0
  aexp_mega_parsec_sqrt3=aexp*mega_parsec*sqrt(3.d0)
  aexp_mega_parsec=aexp*mega_parsec

  if (megaverbose) write(*,*) '-> Match the dynamical haloes hierarchical'// &
 &                 ' structure to the detected one'
  
  allocate(nodedyn(nnodes),touched(nnodes))
  do inode=1,nnodes
     nodedyn(inode)%level=node(inode)%level
     nodedyn(inode)%mother=node(inode)%mother
     nodedyn(inode)%firstchild=node(inode)%firstchild
     nodedyn(inode)%sister=node(inode)%sister
     nodedyn(inode)%oldindex=inode
  enddo

  iteration=0

  Niterationseff=abs(Niterations)
  stopit=.false.
  
  do while (.not.stopit)
  
     if (megaverbose) then
        write(*,*) '-------------------------------------------------'
        write(*,*) 'Iteration :',iteration
        write(*,*) '-------------------------------------------------'
     endif

     if (megaverbose) write(*,*) '-> Compute the properties of the dynamical haloes (averages)'

     do inode=1,nnodes
        nodedyn(inode)%mass=0
        nodedyn(inode)%truemass=0.0d0
        nodedyn(inode)%radius=0.0d0
        nodedyn(inode)%position(1:3)=0.d0
        nodedyn(inode)%velocity(1:3)=0.d0
        touched(inode)=.false.
     enddo
     
     if (megaverbose) write(*,*) '--> 1 : affect each particle to its node'
     incpart=0
     nsubpart=0
     do ipar=1,npart
        ! Substructure in which the particle is
        inode=partnode(ipar)
        if (abs(inode) > 0) nsubpart=nsubpart+1
        if (inode > 0) then
           if (touched(inode)) then
              ! Deal with periodic boundaries
              xpn(1:3)=nodedyn(inode)%position(1:3)/nodedyn(inode)%truemass
              pad(1:3)=pos(1:3,ipar)
              diffpos(1:3)=pad(1:3)-xpn(1:3)
              call deal_with_bound(diffpos(1),pad(1),xlong,xlongs2)
              call deal_with_bound(diffpos(2),pad(2),ylong,ylongs2)
              call deal_with_bound(diffpos(3),pad(3),zlong,zlongs2)
              nodedyn(inode)%position(1:3)=nodedyn(inode)%position(1:3) &
 &                                        +masspart(ipar)*pad(1:3)
           else
              touched(inode)=.true.
              nodedyn(inode)%position(1:3)=masspart(ipar)*pos(1:3,ipar)
           endif
           nodedyn(inode)%velocity(1:3)=nodedyn(inode)%velocity(1:3) &
 &                                     +masspart(ipar)*vel(1:3,ipar)
           nodedyn(inode)%mass=nodedyn(inode)%mass+1
           nodedyn(inode)%truemass=nodedyn(inode)%truemass+masspart(ipar)
           incpart=incpart+1
        endif
     enddo
     if (megaverbose) then
        write(*,*) 'Number of particles in the nodes :',incpart
        write(*,*) 'Number of particles in the nodes + left over particles :',nsubpart
     endif

     ! Complete the calculation of basic dynamical properties of the nodes
     if (megaverbose) write(*,*) '--> 2 : Complete the calculation recursively in the tree'
     call sum_treedyn(1)

     ! periodic boundaries again
     if (megaverbose) write(*,*) '--> 3 : Apply final periodic boundaries (even if not required)'
     inceliminated=0
     do inode=1,nnodes
        if (nodedyn(inode)%mass > 0) then
           nodedyn(inode)%position(1:3)=nodedyn(inode)%position(1:3) &
 &                              /nodedyn(inode)%truemass
           nodedyn(inode)%velocity(1:3)=nodedyn(inode)%velocity(1:3) &
 &                              /nodedyn(inode)%truemass
           call deal_with_bound(nodedyn(inode)%position(1),nodedyn(inode)%position(1),xlong,xlongs2)
           call deal_with_bound(nodedyn(inode)%position(2),nodedyn(inode)%position(2),ylong,ylongs2)
           call deal_with_bound(nodedyn(inode)%position(3),nodedyn(inode)%position(3),zlong,zlongs2)
        else
           write(*,*) 'Elimination of node=',inode
           inceliminated=inceliminated+1
        endif
     enddo

     if (megaverbose) write(*,*) '-> Compute the properties of the dynamical haloes (second moments)'

     ! Compute the least square radius of each node by recursively adding the contribution
     ! of each particle to the node it is contained in and its successive mothers
     do ipar=1,npart
        inode=partnode(ipar)
        if (inode > 0) then
           do while (nodedyn(inode)%level /= 0)
              xpn(1:3)=nodedyn(inode)%position(1:3)
              pad(1:3)=pos(1:3,ipar)
              diffpos(1:3)=pad(1:3)-xpn(1:3)
              call deal_with_bound(diffpos(1),diffpos(1),xlong,xlongs2)
              call deal_with_bound(diffpos(2),diffpos(2),ylong,ylongs2)
              call deal_with_bound(diffpos(3),diffpos(3),zlong,zlongs2)              
              nodedyn(inode)%radius=nodedyn(inode)%radius &
 &                              +masspart(ipar)*(diffpos(1)**2+diffpos(2)**2+diffpos(3)**2)
              inode=nodedyn(inode)%mother
           enddo
        endif
     enddo

     ! Normalize the least square radius of each node
     do inode=1,nnodes
        if (nodedyn(inode)%mass > 0) nodedyn(inode)%radius= &
 &         sqrt(nodedyn(inode)%radius/nodedyn(inode)%truemass)
     enddo

     ! Self consistency check
     if (megaverbose) write(*,*) '-> Check the self-consistency of the results'
     if (iteration == 0) then
        something_wrong=.false.
        do inode=2,nnodes
           if (nodedyn(inode)%mass /= node(inode)%mass) then
              write(*,*) 'inode, masnode, mass=',inode,nodedyn(inode)%mass,node(inode)%mass
              something_wrong=.true.
           endif
        enddo
        if (something_wrong) then
           write(*,*) 'Particle count missmatch in substructures'
           STOP
        endif
        deallocate(node)
     endif

     ! Eliminate nodes with zero mass
     call reconnect_tree

     if ((inceliminated == 0 .and. Niterations < 0 .and. iteration >=1) .or. &
 &       iteration==Niterationseff) stopit=.true.
     iteration=iteration+1

     if (.not.stopit) then

        if (megaverbose) write(*,*) '-> Check if particles are bound to their nodes, starting from the leaves'

        if (megaverbose) write(*,*) '--> 1 : select relevant particles'
        allocate(idspart(nsubpart),idspartnode(nsubpart), &
 &               idspartleve(nsubpart),idsortnode(nsubpart))
        incpart=0
        do ipar=1,npart
           ! Recheck particles that have been previously eliminated
           inode=abs(partnode(ipar))
           if (inode > 0) then
              incpart=incpart+1
              idspart(incpart)=ipar
              idspartnode(incpart)=inode
              idspartleve(incpart)=nodedyn(inode)%level
           endif
        enddo
        nsubpart=incpart
        
        do ipar=1,npart
           partnode(ipar)=0
        enddo
        
        ! Sort the particles by increasing level in the tree
        if (megaverbose) write(*,*) '--> 2 : sort them according to their level in the tree'
        call int_indexx(nsubpart,idspartleve,idsortnode)
        
        incchange=0

        ! Start from particles deepest in the tree and go down
        if (megaverbose) write(*,*) '--> 3 : treat the particles going down from uper to lower level'
        npartleftover=0
        do ispar=nsubpart,1,-1
           ids=idsortnode(ispar)
           inode=idspartnode(ids)
           ipar=idspart(ids)
           ! Check if the particle is bound somewhere to its node
           ! and find the branch (or leave if possible) to which the
           ! the particle is most bound going down into the tree
           isbound=.false.
           call explore_branch(inode,ipar,inodesub,isbound) 
       
           if (isbound) then
              partnode(ipar)=inodesub
              if (inodesub /= inode) incchange=incchange+1
           else
              ! If not go up into the tree untill the particle is bound           
              inodes=inode
              do while (nodedyn(inodes)%level /= 0 .and. .not. isbound)
                 inodefound=inodes
                 call check_bound(ipar,inodefound,isbound,bindingener)
                 inodes=nodedyn(inodes)%mother
              enddo
              
              if (isbound) then
                 ! If the particle is found at some level to be bound, find the 
                 ! branch to which the particle is most bound going down into the tree
                 call explore_branch(inodefound,ipar,inodesub,isbound)
                 partnode(ipar)=inodesub
                 if (inodesub /= inode) incchange=incchange+1
              else
                 ! If the particle is not bound to any structure it means it has to be eliminated
                 npartleftover=npartleftover+1
                 partnode(ipar)=-inodefound
                 if (inodefound /= inode) incchange=incchange+1
              endif
           endif
        enddo
        deallocate(idspart,idspartnode,idspartleve,idsortnode)
        if (megaverbose) then
           write(*,*) 'Number of particles left over in this iteration :',npartleftover
           write(*,*) 'Number of particles which changed host node :',incchange
        endif
     endif
  enddo
end subroutine posttreat

!=======================================================================
subroutine deal_with_bound(x,y,xL,xLs2)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  real(kind=8) :: x,y,xL,xLs2
  if (x >= xLs2) then
     y=y-xL
  elseif (x < -xLs2) then
     y=y+xL
  endif
end subroutine deal_with_bound

!=======================================================================
subroutine reconnect_tree
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inode,idm,ile,idmfcs,idmfc,nsisters
  integer :: incnode,ipar,inodes
  integer,allocatable,dimension(:) :: idnode

  if (megaverbose) write(*,*) '-> Reconnect the tree'

  ! Suppress the nodes with zero mass and reconnect the tree
  do inode=2,nnodes
     if (nodedyn(inode)%mass==0) then
        idm=nodedyn(inode)%mother
        ile=nodedyn(idm)%level
        idmfc=nodedyn(idm)%firstchild
        if (idmfc==inode) then
           nodedyn(idm)%firstchild=nodedyn(inode)%sister
        else
           do while (idmfc /= inode)
              idmfcs=idmfc
              idmfc=nodedyn(idmfc)%sister
           enddo
           nodedyn(idmfcs)%sister=nodedyn(inode)%sister
        endif
     endif
  end do

  ! Due to the reconnexions, some nodes have a single child
  ! This one must be eliminated by setting its mass to zero
  do inode=2,nnodes
     nsisters=0
     idm=nodedyn(inode)%mother
     idmfc=nodedyn(inode)%firstchild
     if (idmfc /= 0) then
        nsisters=1
        do while (idmfc /= 0)
           idmfc=nodedyn(idmfc)%sister
           nsisters=nsisters+1
        enddo
     endif
     if (nsisters==1) then
        nodedyn(nodedyn(inode)%firstchild)%mass=0
        nodedyn(inode)%firstchild=0
     endif
  enddo

  ! Regather the nodes which are not eliminated
  allocate(idnode(0:nnodes))
  incnode=1
  idnode(0)=0
  idnode(1)=1
  do inode=2,nnodes
     if (nodedyn(inode)%mass /= 0) then
        incnode=incnode+1
        idnode(inode)=incnode
     else
        inodes=inode
        do while (nodedyn(inodes)%level /= 1.and.nodedyn(inodes)%mass==0)
           inodes=nodedyn(inodes)%mother
        enddo
        if (nodedyn(inodes)%level==1 .and. nodedyn(inodes)%mass==0) then
           ! particles corresponding to this node definitely eliminated
           ! since the corresponding halo is non dynamically bounded
           idnode(inode)=0 
        else
           ! particles corresponding to this node potentially added 
           ! to the first mother with non zero mass
           idnode(inode)=inodes
        endif
     endif
  enddo
  do inode=2,nnodes
     if (nodedyn(inode)%mass == 0) then
        idnode(inode)=idnode(idnode(inode))
     endif
  enddo

  incnode=1
  do inode=2,nnodes
     if (nodedyn(inode)%mass /= 0) then
        incnode=incnode+1
        nodedyn(incnode)=nodedyn(inode)
     endif
  enddo
  nnodes=incnode

  ! Reestablish the connexions correctly taking into
  ! account the index change of the nodes. 
  do inode=1,nnodes
     nodedyn(inode)%mother=idnode(nodedyn(inode)%mother)
     nodedyn(inode)%sister=idnode(nodedyn(inode)%sister)
     nodedyn(inode)%firstchild=idnode(nodedyn(inode)%firstchild)     
  enddo

  do ipar=1,npart
     if (partnode(ipar) >= 0) then
        partnode(ipar)=idnode(partnode(ipar))
     else
        partnode(ipar)=-idnode(-partnode(ipar))
     endif
  enddo

  deallocate(idnode)
end subroutine reconnect_tree

!=======================================================================
recursive subroutine explore_branch(inodein,ipar,inodesub,isbound)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inodein,ipar,inodesub
  logical :: isbound
  
  real(kind=8) :: bindingener_here,bindingener_t
  logical :: isbound_here
  integer :: inode,inodenext

  call check_bound(ipar,inodein,isbound_here,bindingener_here)
  isbound=isbound_here.or.isbound

  if (isbound_here) then
     inodesub=inodein
     inode=nodedyn(inodein)%firstchild
     inodenext=inode     
     do while (inode /= 0) 
        call check_bound(ipar,inode,isbound_here,bindingener_here)
        if (inode==inodenext) then
           bindingener_t=bindingener_here
        elseif (bindingener_here < bindingener_t) then
           bindingener_t=bindingener_here
           inodenext=inode
        endif
        inode=nodedyn(inode)%sister
     enddo
     if (inodenext /=0) call explore_branch(inodenext,ipar,inodesub,isbound)
  endif

end subroutine explore_branch

!=======================================================================
subroutine check_bound(ipar,inode,isbound,bindingener)
!=======================================================================
! Check if the particle is bounded to the node, assuming this later has
! a isothermal profile rho(r) ~ r^{-2}
! This means that the maximum radius of the halo Rmax is equal to
! sqrt(3) Rsquare, where Rsquare is the mean square radius.
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: ipar,inode
  logical :: isbound
  real(kind=8) :: bindingener

  real(kind=8) :: dv(3),dr(3),kinetic,RR,gmovr

  dr(1:3)=pos(1:3,ipar)-nodedyn(inode)%position(1:3)
  call deal_with_bound(dr(1),dr(1),xlong,xlongs2)
  call deal_with_bound(dr(2),dr(2),ylong,ylongs2)
  call deal_with_bound(dr(3),dr(3),zlong,zlongs2)

  dv(1:3)=vel(1:3,ipar)-nodedyn(inode)%velocity(1:3)+dr(1:3)*aexp_hubble_const
  kinetic=0.5d0*(dv(1)**2+dv(2)**2+dv(3)**2)*convert_in_mps2

  
  RR=sqrt(dr(1)**2+dr(2)**2+dr(3)**2)*aexp_mega_parsec

  gmovr=GMphys*nodedyn(inode)%truemass/max(sqrt(RR),nodedyn(inode)%radius*aexp_mega_parsec_sqrt3)
  bindingener=kinetic-gmovr

  if (bindingener < 0.d0) then
     isbound=.true.
  else
     isbound=.false.
  endif
end subroutine check_bound

!=======================================================================
recursive subroutine sum_treedyn(inodein)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inode,inodein,mass
  real(kind=8) :: tmass,xpn(3),pad(3),diffpos(3)

  inode=nodedyn(inodein)%firstchild
  do while (inode /= 0)
     call sum_treedyn(inode)
     mass=nodedyn(inode)%mass
     if (mass > 0) then
        nodedyn(inodein)%mass=nodedyn(inodein)%mass+mass
        tmass=nodedyn(inode)%truemass

        ! Deal with periodic boundaries
        xpn(1:3)=nodedyn(inodein)%position(1:3)/nodedyn(inodein)%truemass
        pad(1:3)=nodedyn(inode)%position(1:3)/tmass
        diffpos(1:3)=pad(1:3)-xpn(1:3)
        call deal_with_bound(diffpos(1),pad(1),xlong,xlongs2)
        call deal_with_bound(diffpos(2),pad(2),ylong,ylongs2)
        call deal_with_bound(diffpos(3),pad(3),zlong,zlongs2)
        nodedyn(inodein)%position(1:3)=nodedyn(inodein)%position(1:3) &
 &                                    +pad(1:3)*tmass
        nodedyn(inodein)%velocity(1:3)=nodedyn(inodein)%velocity(1:3) &
 &                                    +nodedyn(inode)%velocity(1:3)
        nodedyn(inodein)%truemass=nodedyn(inodein)%truemass+tmass
     endif
     inode=nodedyn(inode)%sister
  enddo
end subroutine sum_treedyn

!=======================================================================
subroutine compute_dist
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: ipar,idist,ip
  integer, allocatable :: iparnei(:,:)
  real(kind=4)  :: poshere(1:3)
  real(kind=4), allocatable :: dist2(:,:)
  real(kind=8)  :: disnorm2,distan,disfac,vxmoy,vymoy,vzmoy,vdis  
  integer :: nblocs,ibloc,ipar1,ipar2,icpu

  if (verbose) write(*,*) 'Compute mean square distance for each particle...'

  allocate(dist2(0:nvoisins,ncpu))
  allocate(iparnei(nvoisins,ncpu))
  allocate(distance(nparbuffer))
  if (velocities) then
     allocate(vxout(nparbuffer),vyout(nparbuffer),vzout(nparbuffer))
     allocate(vdisout(nparbuffer))
  endif

  disnorm2=1.d0/(dble(nvoisins)*max(xlong,ylong,zlong)**2)
  disfac=1./dble(nvoisins+1)

  nblocs=npart/nparbuffer+1
  do ibloc=1,nblocs
     ipar1=(ibloc-1)*nparbuffer+1
     ipar2=min(ibloc*nparbuffer,npart)
!$OMP PARALLEL DO &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(icpu,ipar,distan,idist,vxmoy,vymoy,vzmoy,vdis,ip)
     do icpu=1,ncpu
        do ipar=ipar1+icpu-1,ipar2,ncpu
           call find_nearest_parts(ipar,dist2(0:nvoisins,icpu), &
 &                                 iparnei(0:nvoisins,icpu))
           distan=0.d0
           do idist=1,nvoisins
              distan=distan+dist2(idist,icpu)
           enddo
           distance(ipar-ipar1+1)=sqrt(distan*disnorm2)
           if (velocities) then
              vxmoy=vel(1,ipar)
              vymoy=vel(2,ipar)
              vzmoy=vel(3,ipar)
              vdis=vxmoy**2+vymoy**2+vzmoy**2
              do idist=1,nvoisins
                 ip=iparnei(idist,icpu)
                 vxmoy=vxmoy+vel(1,ip)
                 vymoy=vymoy+vel(2,ip)
                 vzmoy=vzmoy+vel(3,ip)
                 vdis=vdis+vel(1,ip)**2+vel(2,ip)**2+vel(3,ip)**2
              enddo
              vxout(ipar)=vxmoy*disfac
              vyout(ipar)=vymoy*disfac
              vzout(ipar)=vzmoy*disfac
              vdisout(ipar)=sqrt(abs( vdis*disfac-(vxout(ipar)**2 &
 &                               +vyout(ipar)**2+vzout(ipar)**2)))
           endif
        enddo
     enddo
!$OMP END PARALLEL DO
     if (ipar2 >= ipar1) call out_dis(ipar2-ipar1+1)
  enddo
  close(loudis)

  deallocate(dist2)
  deallocate(distance)
  deallocate(iparnei)
  if (velocities) then
     deallocate(vxout,vyout,vzout)
     deallocate(vdisout)
  endif
end subroutine compute_dist

!=======================================================================
subroutine out_dis(npartout)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  logical :: already=.false.
  integer*4 :: nflag,nparmi,nparma,ipar,npartout
  save already

  if (verbose) write(*,*) 'Output the distances in a file...'

  if (.not.already) then
     if (formatted) then
        open(file=filedis,unit=loudis,form='formatted',status='unknown')
     else
        open(file=filedis,unit=loudis,form='unformatted',status='unknown')
     endif
     already=.true.
     if (velocities) then
        nflag=-2
     else
        nflag=-1
     endif
     nparmi=1
     nparma=npart
     if (formatted) then
        write(loudis,*) nflag,npart,nparmi,nparma,nparbuffer
     else
        write(loudis) nflag,npart,nparmi,nparma,nparbuffer
     endif
  endif
  if (formatted) then
     write(loudis,'(5E)') (real(distance(ipar)),ipar=1,npartout)
     if (velocities) then
        write(loudis,'(5E)') (vxout(ipar),ipar=1,npartout)
        write(loudis,'(5E)') (vyout(ipar),ipar=1,npartout)
        write(loudis,'(5E)') (vzout(ipar),ipar=1,npartout)
        write(loudis,'(5E)') (vdisout(ipar),ipar=1,npartout)
     endif
  else
     write(loudis) (distance(ipar),ipar=1,npartout)
     if (velocities) then
        write(loudis) (vxout(ipar),ipar=1,npartout)
        write(loudis) (vyout(ipar),ipar=1,npartout)
        write(loudis) (vzout(ipar),ipar=1,npartout)
        write(loudis) (vdisout(ipar),ipar=1,npartout)
     endif
  endif
end subroutine out_dis

!=======================================================================
subroutine compute_mean_density_and_np
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: ipar,i
  real(kind=4), dimension(0:nvoisins) :: dist2
  integer, dimension(nvoisins) :: iparnei
  real(kind=8)  :: densav,xmasstmp
  logical :: checkaverage=.true.

  if (verbose) write(*,*) 'Compute mean density for each particle...'

  allocate(iparneigh(nhop,npart))
  allocate(density(npart))

  if (verbose) write(*,*) 'First find nearest particles'

!$OMP PARALLEL DO &
!$OMP DEFAULT(SHARED) &
!$OMP PRIVATE(ipar,dist2,iparnei,i,xmasstmp)
  do ipar=1,npart
     call find_nearest_parts(ipar,dist2,iparnei)
     call compute_density(ipar,dist2,iparnei)
     iparneigh(1:nhop,ipar)=iparnei(1:nhop)
     if (megaverbose.and.mod(ipar,10000).eq.0) then
        xmasstmp=0.d0
        do i=1,nvoisins
           xmasstmp=xmasstmp+masspart(iparnei(i))
        enddo
        write(*,*) ipar,density(ipar),real((3./(4.*pi))*  &
 &                  xmasstmp/(dist2(nvoisins)**1.5) &
 &                  *(xlong*ylong*zlong)/dble(npart_true))
     endif
  enddo
!$OMP END PARALLEL DO

! Check for average density
  if (checkaverage) then
     densav=0.d0
     do ipar=1,npart
        densav=(densav*dble(ipar-1)+dble(density(ipar)))/dble(ipar)
     enddo
  endif
  
  if (verbose) write(*,*) 'Average density :',densav

  deallocate(mass_cell)
  deallocate(size_cell)
  deallocate(pos_cell)
  deallocate(sister)
  deallocate(firstchild)

end subroutine compute_mean_density_and_np

!=======================================================================
subroutine out_nei_info
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: ipar,ihop

  if (verbose) write(*,*) &
 &             'output density and nearest neighbourgs in a file...'

  open(unit=lounei,file=filenei,form='unformatted',status='unknown')
  write(lounei) npart,nvoisins,nhop
  write(lounei) (density(ipar),ipar=1,npart)
  do ipar=1,npart
     write(lounei) (iparneigh(ihop,ipar),ihop=1,nhop)
  enddo
  close(lounei)
end subroutine out_nei_info

!=======================================================================
subroutine read_nei_info
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: ipar,ihop,npart_in

  if (verbose) write(*,*) &
 &             'read density and nearest neighbourgs in a file...'

  open(unit=lounei,file=filenei,form='unformatted',status='old')

  read(lounei) npart_in,nvoisins,nhop
  if (npart_in /= npart) then
     write(*,*) 'ERROR in read_nei_info :'
     write(*,*) 'Mismatch in particle count'
     write(*,*) 'npart, npart_in =',npart,npart_in
     STOP
  endif

  allocate(density(npart))
  allocate(iparneigh(nhop,npart))

  read(lounei) density
  do ipar=1,npart
     read(lounei) (iparneigh(ihop,ipar),ihop=1,nhop)
  enddo
  close(lounei)
end subroutine read_nei_info

!=======================================================================
subroutine find_local_maxima
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: ipar,idist,iparid,iparsel,igroup,nmembmax,nmembtot
  real(kind=4)  :: denstest

  if (verbose) write(*,*) 'Now Find local maxima...'

  allocate(igrouppart(npart))
  allocate(idpart(npart))

  ngroups=0
  do ipar=1,npart
     denstest=density(ipar)
     if (denstest.gt.rho_threshold) then
        iparsel=ipar
        do idist=1,nhop
           iparid=iparneigh(idist,ipar)
           if (density(iparid).gt.denstest) then
              iparsel=iparid
              denstest=density(iparid)
           elseif (density(iparid).eq.denstest) then
              iparsel=min(iparsel,iparid)
              if (verbose) &
 &               write(*,*) 'WARNING : equal densities in find_local_maxima.'
           endif
        enddo
        if (iparsel.eq.ipar) then
           ngroups=ngroups+1
           idpart(ipar)=-ngroups
        else
           idpart(ipar)=iparsel
        endif
     endif
  enddo
 
  if (verbose) write(*,*) 'Number of local maxima found :',ngroups

! Now Link the particles associated to the same maximum
  if (verbose) write(*,*) 'Create link list...'

  allocate(densityg(ngroups))
  allocate(nmemb(ngroups))
  allocate(firstpart(ngroups))

  do ipar=1,npart
     if (density(ipar).gt.rho_threshold) then
        iparid=idpart(ipar)
        if (iparid.lt.0) densityg(-iparid)=density(ipar)
        do while (iparid.gt.0)
           iparid=idpart(iparid)
        enddo
        igrouppart(ipar)=-iparid
     else
        igrouppart(ipar)=0
     endif
  enddo

  nmemb(1:ngroups)=0
  firstpart(1:ngroups)=0
  do ipar=1,npart
     igroup=igrouppart(ipar)
     if (igroup.gt.0) then
        idpart(ipar)=firstpart(igroup)   
        firstpart(igroup)=ipar
        nmemb(igroup)=nmemb(igroup)+1
     endif
  enddo

  nmembmax=0
  nmembtot=0
  do igroup=1,ngroups
     nmembmax=max(nmembmax,nmemb(igroup))
     nmembtot=nmembtot+nmemb(igroup)
  enddo

  if (verbose) then
     write(*,*) 'Number of particles of the largest group :',nmembmax
     write(*,*) 'Total number of particles in groups ',nmembtot
  endif
  
end subroutine find_local_maxima

!=======================================================================
subroutine out_partnode
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: i
  
  if (verbose) write(*,*) 'Output the particles to nodes file...'
  
  if (formattedpartnode) then
     open(unit=loupartnode,status='unknown',form='formatted',file=filepartnode)

     write(loupartnode,'(I)') npart
     do i=1,npart
        write(loupartnode,'(I)') partnode(i)
     enddo
     close(loupartnode)
     
  else
     open(unit=loupartnode,status='unknown',form='unformatted',file=filepartnode)

     write(loupartnode) npart
     write(loupartnode) (partnode(i),i=1,npart)
     close(loupartnode)
  endif
end subroutine out_partnode

!=======================================================================
subroutine in_partnode
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: i,npart_in

  if (verbose) write(*,*) 'Read the particles to nodes file...'

  if (formattedinpartnode) then
     open(unit=loupartnode,status='old',form='formatted',file=fileinpartnode)
  
     read(loupartnode,*) npart_in
     if (npart_in /= npart) then
        write(*,*) 'ERROR in in_partnode :'
        write(*,*) 'Missmatch particle count :'
        write(*,*) 'Npart, Npart_in=',Npart,Npart_in
        STOP
     endif
     allocate(partnode(npart))
     read(loupartnode,*) (partnode(i),i=1,npart)
     close(loupartnode)
  else
     open(unit=loupartnode,status='old',form='unformatted',file=fileinpartnode)
  
     read(loupartnode) npart_in
     if (npart_in /= npart) then
        write(*,*) 'ERROR in in_partnode :'
        write(*,*) 'Missmatch particle count :'
        write(*,*) 'Npart, Npart_in=',Npart,Npart_in
        STOP
     endif
     allocate(partnode(npart))
     read(loupartnode) (partnode(i),i=1,npart)
     close(loupartnode)
  endif

end subroutine in_partnode

!=======================================================================
subroutine out_group_tree
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inode,if_simu_units
 
  if (verbose) write(*,*) 'Output the final tree on a file...'

  open(unit=lounode,status='unknown',form='formatted',file=filenode)

  if (simu_units) then
     if_simu_units=1
  else
     if_simu_units=0
  endif
  write(lounode,'(I,E,I,E,E,E,I)') nnodes,real(rho_threshold),nmembthresh,real(alpha), &
 &                             real(fudge),real(fudgepsilon),if_simu_units
  do inode=1,nnodes
     if (simu_units) then
        node(inode)%truemass=node(inode)%truemass/massrenorm
        node(inode)%position(1:3)=(node(inode)%position(1:3) &
 &                            -pos_shift(1:3))/pos_renorm
        node(inode)%radius=node(inode)%radius/pos_renorm
     else
        node(inode)%truemass=node(inode)%truemass*real(mass_in_kg/solar_mass)        
     endif
     write(lounode,'(I,I,I,I,I,I,E,E,E,E,I,E,E,E,E)') &
 &                         inode, &
 &                         node(inode)%level, &
 &                         node(inode)%mother, &
 &                         node(inode)%firstchild, &
 &                         node(inode)%nsisters, &
 &                         node(inode)%sister, &
 &                         node(inode)%rho_saddle, &
 &                         node(inode)%density, &
 &                         node(inode)%densmax, &
 &                         node(inode)%radius, &
 &                         node(inode)%mass, &
 &                         node(inode)%truemass, &
 &                         node(inode)%position(1:3)
  enddo
  close(lounode)
end subroutine out_group_tree

!=======================================================================
subroutine out_group_tree_dyn
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inode,if_simu_units
 
  if (verbose) write(*,*) 'Output the final tree on a file...'

  open(unit=lounodedyn,status='unknown',form='formatted',file=filenodedyn)

  if (simu_units) then
     if_simu_units=1
  else
     if_simu_units=0
  endif
  write(lounodedyn,'(I,E,I,E,E,E,I)') nnodes,real(rho_threshold),nmembthresh,real(alpha), &
 &                             real(fudge),real(fudgepsilon),if_simu_units
  do inode=1,nnodes
     if (simu_units) then
        nodedyn(inode)%truemass=nodedyn(inode)%truemass/massrenorm
        nodedyn(inode)%position(1:3)=(nodedyn(inode)%position(1:3) &
 &                            -pos_shift(1:3))/pos_renorm
        nodedyn(inode)%velocity(1:3)=nodedyn(inode)%velocity(1:3)/velrenorm
        nodedyn(inode)%radius=nodedyn(inode)%radius/pos_renorm
     else
        nodedyn(inode)%truemass=nodedyn(inode)%truemass*mass_in_kg/solar_mass
     endif
     write(lounodedyn,'(I,I,I,I,I,I,E,I,E,E,E,E,E,E,E)') &
 &                         inode, &
 &                         nodedyn(inode)%oldindex, &
 &                         nodedyn(inode)%level, &
 &                         nodedyn(inode)%mother, &
 &                         nodedyn(inode)%firstchild, &
 &                         nodedyn(inode)%sister, &
 &                         real(nodedyn(inode)%radius), &
 &                         nodedyn(inode)%mass, &
 &                         real(nodedyn(inode)%truemass), &
 &                         real(nodedyn(inode)%position(1:3)), &
 &                         real(nodedyn(inode)%velocity(1:3))
  enddo
  close(lounodedyn)
end subroutine out_group_tree_dyn

!=======================================================================
subroutine in_group_tree
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inode,inode_in,if_simu_units

  if (verbose) write(*,*) 'Read the node tree file...'
  
  open(unit=lounode,status='old',form='formatted',file=filenode)

  read(lounode,*) nnodes,rho_threshold,nmembthresh,alpha, &
 &                fudge,epsilon,if_simu_units
  if (if_simu_units==1) then
     simu_units=.true.
  else
     simu_units=.false.
  endif

  allocate(node(nnodes))
  do inode=1,nnodes
     read(lounode,*) inode_in, &
                     node(inode)%level, &
 &                   node(inode)%mother, &
 &                   node(inode)%firstchild, &
 &                   node(inode)%nsisters, &
 &                   node(inode)%sister, &
 &                   node(inode)%rho_saddle, &
 &                   node(inode)%density, &
 &                   node(inode)%densmax, &
 &                   node(inode)%radius, &
 &                   node(inode)%mass, &
 &                   node(inode)%truemass, &
 &                   node(inode)%position(1:3)
  enddo
end subroutine in_group_tree

!=======================================================================
subroutine create_group_tree
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inode,mass,masstmp,igroup,igr1,igr2,igroupref,ipar
  real(kind=8)  :: rhot,posg(3),posref(3),rsquare,densmoy,truemass,truemasstmp

  if (verbose) write(*,*) 'Create the tree of structures of structures'

! End of the branches of the tree
  call compute_saddle_list

  if (verbose) write(*,*) 'Build the hierarchical tree'

  nnodesmax=2*ngroups
! Allocations
  allocate(node(nnodesmax))
  allocate(idgroup(ngroups))
  allocate(color(ngroups))
  allocate(igroupid(ngroups))
  allocate(idgroup_tmp(ngroups))
  if (partinfo) allocate(partnode(npart))

! Initializations
  if (partinfo) then
     do ipar=1,npart
        partnode(ipar)=0
     enddo
  endif

! Iterative loop to build the rest of the tree
  inode=1
  nnodes=1
  rhot=rho_threshold
  node(inode)%mother=0
  mass=0
  truemass=0.d0
  igroupref=0
  do igroup=1,ngroups
     call treat_particles(igroup,rhot,posg,masstmp,igroupref,posref, &
&                          rsquare,densmoy,truemasstmp)
     mass=mass+masstmp
     truemass=truemass+truemasstmp
  enddo

  node(inode)%mass=mass
  node(inode)%truemass=truemass
  node(inode)%radius=0.d0
  node(inode)%density=0.d0
  node(inode)%position(1:3)=0.d0
  node(inode)%densmax=maxval(densityg(1:ngroups))
  node(inode)%rho_saddle=0.
  node(inode)%level=0
  node(inode)%nsisters=0
  node(inode)%sister=0
  igr1=1
  igr2=ngroups
  do igroup=1,ngroups
     idgroup(igroup)=igroup
     igroupid(igroup)=igroup
  enddo
  call create_nodes(rhot,inode,igr1,igr2)

  deallocate(idgroup)
  deallocate(color)
  deallocate(igroupid)
  deallocate(idgroup_tmp)
  deallocate(density)
  deallocate(idpart)
  deallocate(group)
  deallocate(pos)
  if (velocities) deallocate(vel)
end subroutine create_group_tree

!=======================================================================
recursive subroutine create_nodes(rhot,inode,igr1,igr2)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: inode,igr1,igr2
  real(kind=8) :: rhot
  integer :: icolor_select,igroup,icolor,igr,igr_eff
  integer :: inc_color_tot,inode1,mass,masstmp,igroupref
  integer :: inodeout,igr1out,igr2out,isisters,nsisters
  integer :: mass_comp,icolor_ref
  real(kind=8) :: posg(3),posgtmp(3),posref(3),rhotout,rsquaretmp,rsquareg
  real(kind=8) :: posdiffx,posdiffy,posdiffz,densmoyg,densmoytmp,rhoinc,rhot2
  real(kind=8) :: densmoy_comp_max,truemass,truemasstmp
  real(kind=4) :: posfin(3)
  real(kind=4) :: densmaxgroup,densmax 
  integer, allocatable :: igrpos(:),igrinc(:)
  integer, allocatable :: igrposnew(:)
  integer, allocatable :: massg(:)
  real(kind=8),  allocatable :: truemassg(:)
  real(kind=4),  allocatable :: densmaxg(:)
  real(kind=8),  allocatable :: densmoy_comp_maxg(:)
  integer, allocatable :: mass_compg(:)
  real(kind=8),  allocatable :: posgg(:,:)
  real(kind=8),  allocatable :: rsquare(:)
  real(kind=8),  allocatable :: densmoy(:)
  logical, allocatable :: ifok(:)  


  color(igr1:igr2)=0
! Percolate the groups
  icolor_select=0
  do igr=igr1,igr2
     igroup=idgroup(igr)
     if (color(igr).eq.0) then
        icolor_select=icolor_select+1
        call do_colorize(icolor_select,igroup,igr,rhot)
     endif
  enddo

! We select only groups where we are sure of having at least one
! particle above the threshold density rhot
! Then sort them to gather them on the list
  allocate(igrpos(0:icolor_select))
  allocate(igrinc(1:icolor_select))
  igrpos(0)=igr1-1
  igrpos(1:icolor_select)=0
  do igr=igr1,igr2
     icolor=color(igr)
     igroup=idgroup(igr)
     if (densityg(igroup).gt.rhot) &
 &           igrpos(icolor)=igrpos(icolor)+1
  enddo
  do icolor=1,icolor_select
     igrpos(icolor)=igrpos(icolor-1)+igrpos(icolor)
  enddo
  if (igrpos(icolor_select)-igr1+1.eq.0) then
     write(*,*) 'ERROR in create_nodes :'
     write(*,*) 'All subgroups are below the threshold.'
!     STOP
  endif      

  igrinc(1:icolor_select)=0
  do igr=igr1,igr2
     icolor=color(igr)
     igroup=idgroup(igr)
     if (densityg(igroup).gt.rhot) then
        igrinc(icolor)=igrinc(icolor)+1
        igr_eff=igrinc(icolor)+igrpos(icolor-1)
        idgroup_tmp(igr_eff)=igroup
        igroupid(igroup)=igr_eff
     endif
  enddo
  igr2=igrpos(icolor_select)
  idgroup(igr1:igr2)=idgroup_tmp(igr1:igr2)

  inc_color_tot=0
  do icolor=1,icolor_select
     if (igrinc(icolor).gt.0) inc_color_tot=inc_color_tot+1
  enddo 
  allocate(igrposnew(0:inc_color_tot))
  igrposnew(0)=igrpos(0)
  inc_color_tot=0
  do icolor=1,icolor_select
     if (igrinc(icolor).gt.0) then
        inc_color_tot=inc_color_tot+1
        igrposnew(inc_color_tot)=igrpos(icolor)
     endif
  enddo
  deallocate(igrpos)
  deallocate(igrinc)

  isisters=0
  allocate(posgg(1:3,1:inc_color_tot))
  allocate(massg(1:inc_color_tot))
  allocate(truemassg(1:inc_color_tot))
  allocate(densmaxg(1:inc_color_tot))
  allocate(densmoy_comp_maxg(1:inc_color_tot))
  allocate(mass_compg(1:inc_color_tot))
  allocate(rsquare(1:inc_color_tot))
  allocate(densmoy(1:inc_color_tot))
  allocate(ifok(1:inc_color_tot))
  ifok(1:inc_color_tot)=.false.

  do icolor=1,inc_color_tot
     posg(1:3)=0.d0
     mass=0
     truemass=0.d0
     rsquareg=0.d0
     densmoyg=0.d0
     igr1=igrposnew(icolor-1)+1
     igr2=igrposnew(icolor)
     densmaxgroup=-1.
     mass_comp=0
     densmoy_comp_max=-1.
     igroupref=0
     do igr=igr1,igr2
        igroup=idgroup(igr)
        densmaxgroup=max(densmaxgroup,densityg(igroup))
        call treat_particles(igroup,rhot,posgtmp,masstmp, &
 &                           igroupref,posref,rsquaretmp, &
 &                           densmoytmp,truemasstmp)
        posg(1)=posg(1)+posgtmp(1)
        posg(2)=posg(2)+posgtmp(2)
        posg(3)=posg(3)+posgtmp(3)
        rsquareg=rsquareg+rsquaretmp
        mass=mass+masstmp
        truemass=truemass+truemasstmp
        densmoyg=densmoyg+densmoytmp
        densmoytmp=densmoytmp/dble(masstmp)
        mass_comp=max(mass_comp,masstmp)
        if (masstmp > 0) then
           densmoy_comp_max=max(densmoy_comp_max, &
 &             densmoytmp/(1.d0+fudge/sqrt(dble(masstmp))))
        endif
     enddo
     massg(icolor)=mass
     truemassg(icolor)=truemass
     posgg(1:3,icolor)=posg(1:3)
     densmaxg(icolor)=densmaxgroup
     densmoy_comp_maxg(icolor)=densmoy_comp_max
     mass_compg(icolor)=mass_comp
     rsquare(icolor)=sqrt(abs( &
 &             (truemass*rsquareg- &
 &             (posg(1)**2+posg(2)**2+posg(3)**2) )/ &
 &              truemass**2 ))
     densmoy(icolor)=densmoyg/dble(mass)

     ifok(icolor)=mass.ge.nmembthresh.and. &
 &      (densmoy(icolor).gt.rhot*(1.d0+fudge/sqrt(dble(mass))).or. &
 &       densmoy_comp_maxg(icolor).gt.rhot).and. &
 &      densmaxg(icolor).ge.alpha*densmoy(icolor).and. &
 &      rsquare(icolor).ge.epsilon

     if (ifok(icolor)) then
        isisters=isisters+1
        icolor_ref=icolor
     endif
  enddo
  nsisters=isisters
  if (nsisters.gt.1) then
     isisters=0
     inode1=nnodes+1
     do icolor=1,inc_color_tot
        if (ifok(icolor)) then
           isisters=isisters+1
           nnodes=nnodes+1
           if (nnodes.gt.nnodesmax) then
              write(*,*) 'ERROR in create_nodes :'
              write(*,*) 'nnodes > nnodes max'
              STOP
           endif
           if (mod(nnodes,max(nnodesmax/10000,1)).eq.0.and.megaverbose) then
              write(*,*) 'nnodes=',nnodes
           endif
           node(nnodes)%mother=inode
           node(nnodes)%densmax=densmaxg(icolor)
           if (isisters.gt.1) then
              node(nnodes)%sister=nnodes-1
           else
              node(nnodes)%sister=0
           endif
           node(nnodes)%nsisters=nsisters
           node(nnodes)%mass=massg(icolor)
           node(nnodes)%truemass=truemassg(icolor)
           if (mass.eq.0) then
              write(*,*) 'ERROR in create_nodes :'
              write(*,*) 'NULL mass for nnodes=',nnodes
              STOP
           endif
           posfin(1:3)=real(posgg(1:3,icolor)/truemassg(icolor))
           node(nnodes)%radius=real(rsquare(icolor))    
           node(nnodes)%density=real(densmoy(icolor))
           if (posfin(1).ge.xlongs2) then
              posfin(1)=posfin(1)-xlong
           elseif (posfin(1).lt.-xlongs2) then
              posfin(1)=posfin(1)+xlong
           endif
           if (posfin(2).ge.ylongs2) then
              posfin(2)=posfin(2)-ylong
           elseif (posfin(2).lt.-ylongs2) then
              posfin(2)=posfin(2)+ylong
           endif
           if (posfin(3).ge.zlongs2) then
              posfin(3)=posfin(3)-zlong
           elseif (posfin(3).lt.-zlongs2) then
              posfin(3)=posfin(3)+zlong
           endif
           node(nnodes)%position(1:3)=posfin(1:3)
           node(nnodes)%rho_saddle=rhot
           node(nnodes)%level=node(inode)%level+1
           if (megaverbose.and.node(nnodes)%mass.ge.nmembthresh) then
              write(*,*) '*****************************************'
              write(*,*) 'new node :',nnodes
              write(*,*) 'level    :',node(nnodes)%level
              write(*,*) 'nsisters :',node(nnodes)%nsisters
              write(*,*) 'mass     :',node(nnodes)%mass
              write(*,*) 'true mass:',node(nnodes)%truemass
              write(*,*) 'radius   :',node(nnodes)%radius
              write(*,*) 'position :',node(nnodes)%position
              write(*,*) 'rho_saddl:',node(nnodes)%rho_saddle
              write(*,*) 'rhomax   :',node(nnodes)%densmax
              write(*,*) '*****************************************'
           endif
        endif
     enddo
     node(inode)%firstchild=nnodes
     inodeout=inode1
     do icolor=1,inc_color_tot
        if (ifok(icolor)) then
           igr1out=igrposnew(icolor-1)+1
           igr2out=igrposnew(icolor)
           if (partinfo) then
              do igr=igr1out,igr2out
                 call paint_particles(idgroup(igr),inodeout,rhot)
              enddo
           endif
           rhotout=rhot*(1.d0+fudge/sqrt(dble(mass_compg(icolor))))
           if (igr2out.ne.igr1out) then
              call create_nodes(rhotout,inodeout,igr1out,igr2out)
           else
              node(inodeout)%firstchild=0
           endif
           inodeout=inodeout+1
        endif
     enddo
  elseif (nsisters.eq.1) then
     inodeout=inode
     rhotout=rhot*(1.d0+fudge/sqrt(dble(mass_compg(icolor_ref))))
     igr1out=igrposnew(0)+1
     igr2out=igrposnew(inc_color_tot)
     if (igr2out.ne.igr1out) then
        call create_nodes(rhotout,inodeout,igr1out,igr2out)
     else
        node(inode)%firstchild=0
     endif
  else
     node(inode)%firstchild=0
  endif
  deallocate(igrposnew)
  deallocate(posgg)
  deallocate(massg)
  deallocate(truemassg)
  deallocate(densmaxg)
  deallocate(densmoy_comp_maxg)
  deallocate(densmoy)
  deallocate(mass_compg)
  deallocate(rsquare)
  deallocate(ifok)

end subroutine create_nodes

!=======================================================================
subroutine paint_particles(igroup,inode,rhot)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: igroup,inode
  real(kind=8) :: rhot

  integer :: ipar

  ipar=firstpart(igroup)
  do while (ipar.gt.0)
     if (density(ipar).gt.rhot) then
        partnode(ipar)=inode
     endif
     ipar=idpart(ipar)
  enddo
end subroutine paint_particles

!=======================================================================
subroutine treat_particles(igroup,rhot,posg,mass,igroupref,posref, &
&                          rsquare,densmoy,truemass)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  real(kind=8) :: rhot
  real(kind=8) :: posg(3),posref(3)
  real(kind=8) :: posdiffx,posdiffy,posdiffz,rsquare,densmoy,truemass,xmasspart
  real(kind=4) :: densmax,densmin
  integer :: mass,ipar,iparold,igroup,igroupref
  logical :: first_good

  mass=0
  truemass=0.d0
  rsquare=0.d0
  densmoy=0.d0
  posg(1:3)=0.d0
  ipar=firstpart(igroup)
  first_good=.false.
  do while (ipar.gt.0)
     if (density(ipar).gt.rhot) then
        if (.not.first_good) then
           if (igroupref.eq.0) then
              posref(1:3)=dble(pos(1:3,ipar))
              igroupref=igroup
           endif
           first_good=.true.
           firstpart(igroup)=ipar
           densmin=density(ipar)
           densmax=densmin
        else
           idpart(iparold)=ipar
        endif
        iparold=ipar
        mass=mass+1
        xmasspart=masspart(ipar)
        truemass=truemass+xmasspart
        posdiffx=dble(pos(1,ipar))-posref(1)
        posdiffy=dble(pos(2,ipar))-posref(2)
        posdiffz=dble(pos(3,ipar))-posref(3)
        if (posdiffx.ge.xlongs2) then
           posdiffx=posdiffx-xlong
        elseif (posdiffx.lt.-xlongs2) then
           posdiffx=posdiffx+xlong
        endif
        if (posdiffy.ge.ylongs2) then
           posdiffy=posdiffy-ylong
        elseif (posdiffy.lt.-ylongs2) then
           posdiffy=posdiffy+ylong
        endif
        if (posdiffz.ge.zlongs2) then
           posdiffz=posdiffz-zlong
        elseif (posdiffz.lt.-zlongs2) then
           posdiffz=posdiffz+zlong
        endif
        posdiffx=posdiffx+posref(1)
        posdiffy=posdiffy+posref(2)
        posdiffz=posdiffz+posref(3)
        posg(1)=posg(1)+posdiffx*xmasspart
        posg(2)=posg(2)+posdiffy*xmasspart
        posg(3)=posg(3)+posdiffz*xmasspart
        rsquare=rsquare+xmasspart*(posdiffx**2+posdiffy**2+posdiffz**2)
        densmoy=densmoy+dble(density(ipar))
        densmax=max(densmax,density(ipar))
        densmin=min(densmin,density(ipar))
     endif
     ipar=idpart(ipar)
  enddo
  if (.not.first_good) firstpart(igroup)=0

  if (densmin.le.rhot.or.densmax.ne.densityg(igroup)) then
     write(*,*) 'ERROR in treat_particles'
     write(*,*) 'igroup, densmax, rhot=',igroup,densityg(igroup),rhot
     write(*,*) 'denslow, denshigh    =',densmin,densmax
     STOP
  endif

end subroutine treat_particles

!=======================================================================
recursive subroutine do_colorize(icolor_select,igroup,igr,rhot)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: icolor_select,igroup,igr
  integer :: ineig,igroup2,igr2,neig
  real(kind=8)  :: rhot  

  color(igr)=icolor_select
  neig=group(igroup)%nhnei
  do ineig=1,group(igroup)%nhnei
     if (group(igroup)%rho_saddle_gr(ineig).gt.rhot) then
! We connect this group to its neighbourg
        igroup2=group(igroup)%isad_gr(ineig)
        igr2=igroupid(igroup2)
        if (color(igr2).eq.0) then
           call do_colorize(icolor_select,igroup2,igr2,rhot) 
        elseif (color(igr2).ne.icolor_select) then
           write(*,*) 'ERROR in do_colorize : color(igr2) <> icolor_select'
           write(*,*) 'The connections are not symmetric.'
           STOP
        endif
     else
! We do not need this saddle anymore (and use the fact that saddles 
! are ranked in decreasing order)
        neig=neig-1
     endif
  enddo
  group(igroup)%nhnei=neig

end subroutine do_colorize

!=======================================================================
subroutine compute_saddle_list
!=======================================================================
! Compute the lowest density threshold below which each group is 
! connected to an other one
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: igroup,ipar1,ipar2,igroup2,ineig,idist,igroup1,ineig2
  integer :: neig,ineigal,isister,in1,in2,idestroy,isaddle,icon_count
  integer :: incgroup,i
  real(kind=4)  :: density1,density2,rho_sad12,rho_sad_glob
  logical :: exist
  logical, allocatable :: touch(:)
  integer, allocatable :: listg(:)
  real(kind=4),  allocatable :: rho_sad(:)  
  integer, allocatable :: isad(:)
  integer, allocatable :: indx(:)

  if (verbose) write(*,*) 'Fill the end of the branches of the group tree'

! Allocate the array of nodes
  allocate(group(ngroups))

  if (verbose) write(*,*) 'First count the number of neighbourgs'// &
&                         ' of each elementary group...'

  allocate(touch(ngroups))
  allocate(listg(ngroups))

  touch(1:ngroups)=.false.


! First count the number of neighbourgs for each group to allocate
! arrays isad_in,isad_out,rho_saddle_gr
  do igroup1=1,ngroups
     ineig=0
     ipar1=firstpart(igroup1)
! Loop on all the members of the group
     do while (ipar1.gt.0)
        do idist=1,nhop
           ipar2=iparneigh(idist,ipar1)
           igroup2=igrouppart(ipar2)
! we test that we are in a group (i.e. that density(ipar) >= rho_threshold)
! and that this group is different from the one we are sitting on
           if (igroup2.gt.0.and.igroup2.ne.igroup1) then
              if (.not.touch(igroup2)) then
                 ineig=ineig+1
                 touch(igroup2)=.true.
                 listg(ineig)=igroup2
              endif
           endif
        enddo
! Next member
        ipar1=idpart(ipar1)
     enddo
! Reinitialize touch
     do in1=1,ineig
        igroup2=listg(in1)
        touch(igroup2)=.false.
     enddo
! Allocate the nodes 
     group(igroup1)%nhnei=ineig
     ineigal=max(ineig,1)
     allocate(group(igroup1)%isad_gr(1:ineigal))
     allocate(group(igroup1)%rho_saddle_gr(1:ineigal))
  enddo 

  if (verbose) write(*,*) 'Compute lists of neighbourgs and saddle points...'


! arrays isad_in,isad_out,rho_saddle_gr
  do igroup1=1,ngroups
! No calculation necessary if no neighbourg
     neig=group(igroup1)%nhnei
     if (neig.gt.0) then
        ineig=0
        ipar1=firstpart(igroup1)
        allocate(rho_sad(1:neig))
! Loop on all the members of the group
        do while (ipar1.gt.0)
           density1=density(ipar1)
           do idist=1,nhop
              ipar2=iparneigh(idist,ipar1)
              igroup2=igrouppart(ipar2)
! we test that we are in a group (i.e. that density(ipar) >= rho_threshold)
! and that this group is different from the one we are sitting on
              if (igroup2.gt.0.and.igroup2.ne.igroup1) then
                 density2=density(ipar2)
                 if (.not.touch(igroup2)) then
                    ineig=ineig+1
                    touch(igroup2)=.true.
                    listg(igroup2)=ineig
                    rho_sad12=min(density1,density2)
                    rho_sad(ineig)=rho_sad12
                    group(igroup1)%isad_gr(ineig)=igroup2
                 else
                    ineig2=listg(igroup2)
                    rho_sad12=min(density1,density2)
                    rho_sad(ineig2)=max(rho_sad(ineig2),rho_sad12)
                 endif
              endif
           enddo
! Next member
           ipar1=idpart(ipar1)
        enddo
        if (ineig.ne.neig) then
! Consistency checking
           write(*,*) 'ERROR in compute_saddle_list :'
           write(*,*) 'The number of neighbourgs does not match.'
           write(*,*) 'ineig, neig =',ineig,neig
           STOP
        endif
        group(igroup1)%rho_saddle_gr(1:ineig)=rho_sad(1:ineig)
        deallocate(rho_sad)
! Reinitialize touch
        do in1=1,ineig
           igroup2=group(igroup1)%isad_gr(in1)
           touch(igroup2)=.false.
        enddo
! No neighbourg
     endif
  enddo 

  deallocate(touch)
  deallocate(listg)

  if (verbose) write(*,*) 'Establish symmetry in connections...'

! Total number of connections count
  icon_count=0

! Destroy the connections between 2 groups which are not symmetric
! This might be rather slow and might be discarded later
  idestroy=0
  do igroup1=1,ngroups
     if (group(igroup1)%nhnei.gt.0) then
        do in1=1,group(igroup1)%nhnei
           exist=.false.
           igroup2=group(igroup1)%isad_gr(in1)
           if (igroup2.gt.0) then
              do in2=1,group(igroup2)%nhnei
                 if (group(igroup2)%isad_gr(in2).eq.igroup1) then
                    exist=.true.
                    rho_sad12=min(group(igroup2)%rho_saddle_gr(in2), &
 &                                group(igroup1)%rho_saddle_gr(in1))
                    group(igroup2)%rho_saddle_gr(in2)=rho_sad12
                    group(igroup1)%rho_saddle_gr(in1)=rho_sad12
                 endif
              enddo
           endif
           if (.not.exist) then
              group(igroup1)%isad_gr(in1)=0
              idestroy=idestroy+1
           else
              icon_count=icon_count+1
           endif
        enddo
     endif
  enddo

  if (verbose) write(*,*) 'Number of connections removed :',idestroy
  if (verbose) write(*,*) 'Total number of connections remaining :',icon_count

  if (verbose) then
     write(*,*) 'Rebuild groups with undesired connections removed...'
  endif


! Rebuild the group list correspondingly with the connections removed
! And sort the list of saddle points
  do igroup1=1,ngroups
     neig=group(igroup1)%nhnei
     if (neig.gt.0) then
        allocate(rho_sad(neig))
        allocate(isad(neig))
        ineig=0
        do in1=1,neig
           igroup2=group(igroup1)%isad_gr(in1)
           if (igroup2.gt.0) then
              ineig=ineig+1
              rho_sad(ineig)=group(igroup1)%rho_saddle_gr(in1)
              isad(ineig)=igroup2
           endif
        enddo
        deallocate(group(igroup1)%isad_gr)
        deallocate(group(igroup1)%rho_saddle_gr)
        ineigal=max(ineig,1)
        allocate(group(igroup1)%isad_gr(ineigal))
        allocate(group(igroup1)%rho_saddle_gr(ineigal))
        group(igroup1)%nhnei=ineig
        if (ineig.gt.0) then
! sort the saddle points by decreasing order
           allocate(indx(ineig))
           call indexx(ineig,rho_sad,indx)
           do i=1,ineig
              ineig2=indx(i)
              group(igroup1)%isad_gr(ineig-i+1)=isad(ineig2)
              group(igroup1)%rho_saddle_gr(ineig-i+1)=rho_sad(ineig2)
           enddo
           deallocate(indx)
        endif
        deallocate(rho_sad)
        deallocate(isad)
     endif
  enddo

  deallocate(iparneigh)
  deallocate(igrouppart)

end subroutine compute_saddle_list

!=======================================================================
subroutine compute_density(ipar,dist2,iparnei)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  real(kind=4)  :: dist2(0:nvoisins)
  integer :: iparnei(nvoisins)
  real(kind=4)  :: r,unsr,contrib,spline
  integer :: idist,ipar

  r=sqrt(dist2(nvoisins))*0.5
  unsr=1./r
  contrib=0.
  do idist=1,nvoisins-1
     contrib=contrib+masspart(iparnei(idist))*spline(sqrt(dist2(idist))*unsr)
  enddo
! Add the contribution of the particle itself and normalize properly
! to get a density with average unity (if computed on an uniform grid)
  density(ipar)=(xlong*ylong*zlong)*(contrib+masspart(ipar)) &
 &              /((pi*r**3)*float(npart_true))
end subroutine compute_density

!=======================================================================
function spline(x)
!=======================================================================
  implicit none
  real(kind=4) :: spline,x

  if (x.le.1.) then
     spline=1.-1.5*x**2+0.75*x**3
  elseif (x.le.2.) then
     spline=0.25*(2.-x)**3
  else
     spline=0.
  endif
end function spline

!=======================================================================
subroutine find_nearest_parts(ipar,dist2,iparnei)
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: ipar,idist,icellid,inccellpart
  real(kind=4)  :: dist2(0:nvoisins)
  integer :: iparnei(nvoisins)
  real(kind=4)  :: poshere(1:3)

  poshere(1:3)=pos(1:3,ipar)
  dist2(0)=0.
  do idist=1,nvoisins
     dist2(idist)=bignum
  enddo
  icellid=1
  inccellpart=0
  call walk_tree(mass_cell,size_cell,pos_cell,sister, &
 &               firstchild,ncellmx,icellid,poshere,dist2, &
 &               nvoisins,real(xlong),real(ylong),real(zlong), &
 &               ipar,inccellpart,iparnei,pos,npart,idpart)
end subroutine find_nearest_parts

!=======================================================================
recursive subroutine walk_tree(mass_cell,size_cell,pos_cell,sister, &
 &                   firstchild,ncellmx,icellidin,poshere,dist2,nvoisins, &
 &                   xlong,ylong,zlong,iparid,inccellpart,iparnei,pos,npart,idpart)
!=======================================================================
  implicit none
  integer :: ncellmx,nvoisins,npart
  integer :: mass_cell(ncellmx),idpart(npart)
  real(kind=4)  :: size_cell(ncellmx)
  real(kind=4)  :: pos_cell(3,ncellmx)
  real(kind=4)  :: pos(3,npart)
  integer :: sister(ncellmx),firstchild(ncellmx)
  integer :: icellidin,icellid,iparid,inccellpart,ic,iparcell
  real(kind=4)  :: poshere(3),dist2(0:nvoisins)
  real(kind=4)  :: xlong,ylong,zlong
  real(kind=4)  :: dx,dy,dz,distance2,sc
  integer :: idist,inc
  integer :: icellid_out
  real(kind=4)  :: discell2(0:8)
  integer :: iparnei(nvoisins)
  integer :: icid(8)

  integer :: i,npart_pos_this_node
  real(kind=4) :: distance2p

  icellid=firstchild(icellidin)
  inc=1
  discell2(0)=0
  discell2(1:8)=1.e30
  do while (icellid.ne.0)
     sc=size_cell(icellid)
     dx=abs(pos_cell(1,icellid)-poshere(1))
     dx=max(0.,min(dx,xlong-dx)-sc)
     dy=abs(pos_cell(2,icellid)-poshere(2))
     dy=max(0.,min(dy,ylong-dy)-sc)
     dz=abs(pos_cell(3,icellid)-poshere(3))
     dz=max(0.,min(dz,zlong-dz)-sc)
     distance2=dx**2+dy**2+dz**2
     if (distance2.lt.dist2(nvoisins)) then
        idist=inc-1
        do while (discell2(idist).gt.distance2)
           discell2(idist+1)=discell2(idist)
           icid(idist+1)=icid(idist)
           idist=idist-1
        enddo
        discell2(idist+1)=distance2
        icid(idist+1)=icellid
        inc=inc+1
     endif
     icellid=sister(icellid)
  enddo
  inccellpart=inccellpart+inc-1
  do ic=1,inc-1
     icellid_out=icid(ic)
     if (firstchild(icellid_out) < 0) then
        if (discell2(ic).lt.dist2(nvoisins)) then
           npart_pos_this_node=-firstchild(icellid_out)-1
           do i=npart_pos_this_node+1,npart_pos_this_node+mass_cell(icellid_out)
              iparcell=idpart(i)
              dx=abs(pos(1,iparcell)-poshere(1))
              dx=max(0.,min(dx,xlong-dx))
              dy=abs(pos(2,iparcell)-poshere(2))
              dy=max(0.,min(dy,ylong-dy))
              dz=abs(pos(3,iparcell)-poshere(3))
              dz=max(0.,min(dz,zlong-dz))
              distance2p=dx**2+dy**2+dz**2
              if (distance2p .lt. dist2(nvoisins)) then 
                 if (iparcell.ne.iparid) then
                    idist=nvoisins-1
                    do while (dist2(idist).gt.distance2p)
                       dist2(idist+1)=dist2(idist)
                       iparnei(idist+1)=iparnei(idist)
                       idist=idist-1
                    enddo
                    dist2(idist+1)=distance2p
                    iparnei(idist+1)=iparcell
                 endif
              endif
           enddo
        endif              
     elseif (discell2(ic).lt.dist2(nvoisins)) then
        call walk_tree(mass_cell,size_cell,pos_cell,sister, &
 &                   firstchild,ncellmx,icellid_out,poshere,dist2, &
 &                   nvoisins,xlong,ylong,zlong,iparid, &
 &                   inccellpart,iparnei,pos,npart,idpart)
     endif
  enddo
end subroutine walk_tree

!=======================================================================
subroutine create_tree_structure
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  integer :: nlevel,inccell,idmother,ipar
  integer :: npart_this_node,npart_pos_this_node
  integer :: ncell
  real(kind=4) :: pos_this_node(3)

  if (verbose) write(*,*) 'Create tree structure...'

  ncellmx=2*npart
  ncellbuffer=max(nint(0.1*npart),ncellbuffermin)
  allocate(idpart(npart))
  allocate(idpart_tmp(npart))
  allocate(mass_cell(ncellmx))
  allocate(size_cell(ncellmx))
  allocate(pos_cell(3,ncellmx))
  allocate(sister(1:ncellmx))
  allocate(firstchild(1:ncellmx))
  do ipar=1,npart
     idpart(ipar)=ipar
  enddo
  nlevel=0
  inccell=0
  idmother=0
  pos_this_node(1:3)=0.
  npart_this_node=npart
  npart_pos_this_node=0
  idpart_tmp(1:npart)=0
  pos_cell(1:3,1:ncellmx)=0.
  size_cell(1:ncellmx)=0.
  mass_cell(1:ncellmx)=0
  sister(1:ncellmx)=0
  firstchild(1:ncellmx)=0
  sizeroot=real(max(xlong,ylong,zlong))
    
  call create_KDtree(nlevel,pos_this_node, &
 &                   npart_this_node,npart_pos_this_node,inccell, &
 &                   idmother)
  ncell=inccell

  if (verbose) write(*,*) 'total number of cells =',ncell

  deallocate(idpart_tmp)
end subroutine create_tree_structure


!=======================================================================
recursive subroutine create_KDtree(nlevel,pos_this_node,npart_this_node, &
 &                npart_pos_this_node,inccell,idmother)
!=======================================================================
!  nlevel : level of the node in the octree. Level zero corresponds to 
!           the full box
!  pos_this_node : position of the center of this node
!  npart  : total number of particles 
!  idpart : array of dimension npart containing the id of each
!           particle. It is sorted such that neighboring particles in
!           this array belong to the same cell node.
!  idpart_tmp : temporary array of same size used as a buffer to sort
!           idpart.
!  npart_this_node : number of particles in the considered node
!  npart_pos_this_node : first position in idpart of the particles 
!           belonging to this node
!  pos :    array of dimension 3.npart giving the positions of 
!           each particle belonging to the halo
!  inccell : cell id number for the newly created structured grid site
!  pos_cell : array of dimension 3.npart (at most) corresponding
!           to the positions of each node of the structured grid
!  mass_cell : array of dimension npart (at most) giving the 
!           number of particles in each node of the structured grid
!  size_cell : array of dimension npart (at most) giving half the
!           size of the cube forming each node of the structured grid
!  sizeroot : size of the root cell (nlevel=0)
!  idmother : id of the mother cell
!  sister   : sister of a cell (at the same level, with the same mother)
!  firstchild : first child of a cell (then the next ones are found 
!             with the array sister). If it is a cell containing only
!           one particle, it gives the id of the particle.
!  ncellmx : maximum number of cells
!  megaverbose : detailed verbose mode
!=======================================================================
   use compute_neiKDtree_common
   implicit none

   integer :: nlevel,npart_pos_this_node,npart_this_node
   real(kind=4)  :: pos_ref(3,0:7)
   real(kind=4)  :: pos_this_node(3)   
   integer :: ipar,icid,icellid,j,inccell,incsubcellmin,nlevel_out
   integer :: npart_pos_this_node_out,npart_this_node_out
   integer :: incsubcell(0:7),nsubcell(0:7)
   real(kind=4)  :: xtest(3),pos_this_node_out(3)
   integer :: idmother,idmother_out,idp

   integer :: ncellmx_old
   integer, allocatable :: mass_cell_tmp(:),sister_tmp(:),firstchild_tmp(:)
   real(kind=4), allocatable :: size_cell_tmp(:),pos_cell_tmp(:,:)

!  pos_ref : an array used to find positions of the 8 subcells in this
!           node.
   data  pos_ref /-1.,-1.,-1., &
 &                 1.,-1.,-1., &
 &                -1., 1.,-1., &
 &                 1., 1.,-1., &
 &                -1.,-1., 1., &
 &                 1.,-1., 1., &
 &                -1., 1., 1., &
 &                 1., 1., 1.  / 


   if (npart_this_node.gt.0) then
      inccell=inccell+1
      if (mod(inccell,1000000).eq.0.and.megaverbose) write(*,*) 'inccell=',inccell
      if (inccell.gt.ncellmx) then
         ncellmx_old=ncellmx
         ncellmx=ncellmx+ncellbuffer
         if (megaverbose) write(*,*) &
 &          'ncellmx is too small. Increase it and reallocate arrays accordingly'
         allocate(mass_cell_tmp(ncellmx_old))
         mass_cell_tmp(1:ncellmx_old)=mass_cell(1:ncellmx_old)
         deallocate(mass_cell)
         allocate(mass_cell(ncellmx))
         mass_cell(1:ncellmx_old)=mass_cell_tmp(1:ncellmx_old)
         deallocate(mass_cell_tmp)
         allocate(sister_tmp(ncellmx_old))
         sister_tmp(1:ncellmx_old)=sister(1:ncellmx_old)
         deallocate(sister)
         allocate(sister(ncellmx))
         sister(1:ncellmx_old)=sister_tmp(1:ncellmx_old)
         deallocate(sister_tmp)
         allocate(firstchild_tmp(ncellmx_old))
         firstchild_tmp(1:ncellmx_old)=firstchild(1:ncellmx_old)
         deallocate(firstchild)
         allocate(firstchild(ncellmx))
         firstchild(1:ncellmx_old)=firstchild_tmp(1:ncellmx_old)
         firstchild(ncellmx_old:ncellmx)=0
         deallocate(firstchild_tmp)
         allocate(size_cell_tmp(ncellmx_old))
         size_cell_tmp(1:ncellmx_old)=size_cell(1:ncellmx_old)
         deallocate(size_cell)
         allocate(size_cell(ncellmx))
         size_cell(1:ncellmx_old)=size_cell_tmp(1:ncellmx_old)
         deallocate(size_cell_tmp)
         allocate(pos_cell_tmp(3,ncellmx_old))
         pos_cell_tmp(1:3,1:ncellmx_old)=pos_cell(1:3,1:ncellmx_old)
         deallocate(pos_cell)
         allocate(pos_cell(3,ncellmx))
         pos_cell(1:3,1:ncellmx_old)=pos_cell_tmp(1:3,1:ncellmx_old)
         deallocate(pos_cell_tmp)
      endif
      pos_cell(1:3,inccell)=pos_this_node(1:3)
      mass_cell(inccell)=npart_this_node
      size_cell(inccell)=2.**(-nlevel)*sizeroot*0.5
      if (idmother.gt.0) then
         sister(inccell)=firstchild(idmother)
         firstchild(idmother)=inccell
      endif
!     If there is only one particle in the node or we have reach
!     maximum level of refinement, we are done
      if ((npart_this_node <= npartpercell).or.(nlevel.eq.nlevelmax)) then
         firstchild(inccell)=-(npart_pos_this_node+1)
         return
      endif
   else
      return
   endif


!  Count the number of particles in each subcell of this node
   incsubcell(0:7)=0
   do ipar=npart_pos_this_node+1,npart_pos_this_node+npart_this_node
      xtest(1:3)=pos(1:3,idpart(ipar))-pos_this_node(1:3)
      icid=icellid(xtest)
      incsubcell(icid)=incsubcell(icid)+1
   enddo

!  Create the array of positions of the first particle of the lists
!  of particles belonging to each subnode
   nsubcell(0)=0
   do j=1,7
      nsubcell(j)=nsubcell(j-1)+incsubcell(j-1)
   enddo

!  Sort the array of ids (idpart) to gather the particles belonging
!  to the same subnode. Put the result in idpart_tmp.
   incsubcell(0:7)=0
   do ipar=npart_pos_this_node+1,npart_pos_this_node+npart_this_node
      xtest(1:3)=pos(1:3,idpart(ipar))-pos_this_node(1:3)
      icid=icellid(xtest)
      incsubcell(icid)=incsubcell(icid)+1
      idpart_tmp(incsubcell(icid)+nsubcell(icid)+npart_pos_this_node)=idpart(ipar)
   enddo
   
!  Put back the sorted ids in idpart
   do ipar=npart_pos_this_node+1,npart_pos_this_node+npart_this_node
      idpart(ipar)=idpart_tmp(ipar)
   enddo

!  Call again the routine for the 8 subnodes:
!  Compute positions of subnodes, new level of refinement, 
!  positions in the array idpart corresponding to the subnodes,
!  and call for the treatment recursively.
   nlevel_out=nlevel+1
   idmother_out=inccell
   do j=0,7
      pos_this_node_out(1:3)=pos_this_node(1:3)+ &
 &                           sizeroot*pos_ref(1:3,j)*2.**(-nlevel-2)
      npart_pos_this_node_out=npart_pos_this_node+nsubcell(j)
      npart_this_node_out=incsubcell(j)
      call create_KDtree(nlevel_out,pos_this_node_out,npart_this_node_out, &
 &                npart_pos_this_node_out,inccell,idmother_out)
   enddo
end subroutine create_KDtree

!=======================================================================
function icellid(xtest)
!=======================================================================
!  Compute cell id corresponding to the signs of coordinates of xtest
!  as follows :
!  (-,-,-) : 0
!  (+,-,-) : 1
!  (-,+,-) : 2
!  (+,+,-) : 3
!  (-,-,+) : 4
!  (+,-,+) : 5
!  (-,+,+) : 6
!  (+,+,+) : 7
!  For self-consistency, the array pos_ref should be defined exactly 
!  with the same conventions
!=======================================================================
   implicit none
   integer :: icellid,j,icellid3d(3)
   real(kind=4) :: xtest(3)
   do j=1,3
      if (xtest(j).ge.0) then
         icellid3d(j)=1
      else
         icellid3d(j)=0
      endif
   enddo
   icellid=icellid3d(1)+2*icellid3d(2)+4*icellid3d(3)
end function icellid


!=======================================================================
subroutine read_part
!=======================================================================
!
! What we really need is
!
! HEADER QUANTITIES
! -----------------
!
! Npart : the number of particles in the input file
! hubble : H0/100 where H0 is the present time Hubble constant in km/s/Mpc
! Omega0 : density parameter
! OmegaL : cosmological constant
! aexp   : expansion factor normalized to unity at present time
! mass_in_kg : multiplicative factor to obtain the masses in kg
! xlong, ylong, zlong : size of the box in Mpc (if possible, otherwise
!          action=posttreat not possible)
! xlongs2, ylongs2, zlongs2 : xlong/2, ylong/2, zlong/2
!
! POSITIONS :
! -----------
!
! ((pos(i,j),i=1,3),j=1,npart) : particles coordinates in Mpc
!          -xlongs2 <= pos(1,:) < xlongs2
!          -ylongs2 <= pos(2,:) < ylongs2
!          -zlongs2 <= pos(3,:) < zlongs2
!
! MASSES :
! --------
!
! (mass(i),i=1,npart) : masse of each particle, which multiplied by
!          mass_in_kg is then in kg.
!
! VELOCITIES :
! ------------
!
! if (velocities) :
! ((vel(i,j),i=1,3),j=1,npart) : particles velocities in km/s (if
!          possible, otherwise action=posttreat not possible)
!
! MISC :
! ------
! pos_shift(1:3), pos_renorm, massrenorm, velrenorm : corrective terms to
!          reobtain units of the code for the output of the nodes in
!          the case action=adaptahop
! epsilon : estimate of the spatial resolution of the simulation in Mpc
!
!=======================================================================
  use compute_neiKDtree_common
  implicit none

! Used by several formats
  integer :: i,j,npartp,impi
  character*255 :: filempi

! PM : Ntype=0
  character*64 :: alphalab
  real(kind=4) :: alab(40),aexplu,vmax,dt,wcin,wpot,rjns
  real(kind=8) :: fnxlu,fnylu,fnzlu,xmult,vmult
  real(kind=8) :: fnxfac,fnyfac,fnzfac
  real(kind=8) :: amax
  real(kind=4),allocatable :: tmp(:)
  integer :: itime

! Gadget : Ntype=1 
  integer :: npartold
  integer :: npart_in(0:5), nall_in(0:5)
  real(kind=8) :: massarr_in(0:5), a_in, redshift_in
  integer :: unused_in(64-6-12-4-2*4-6*4-2*4-4*8)
  integer :: flag_sfr_in,flag_feedback_in,flag_cooling_in,numfiles_in
  real(kind=8) :: xLbox_in
  real(kind=8) :: facco

! RAMSES unformatted cosmological multiple MPI files : Ntype=2
  character*6 :: string
  integer :: ndimp,ncpup
  real(kind=8) :: mymasstot
  real(kind=4),allocatable :: tmpmpi(:)
  
! CONES : Ntype=3
  character*255 :: filecone
  integer :: Nfiles,ifile,inc
  real(kind=4) :: hubblein,omega0in,omegaLin,zmaxconein
  real(kind=4) :: the1conein,the2conein,thetashiftin,phishiftin,padin
  real(kind=4) :: xin,yin,zin,vxin,vyin,vzin,redshiftin,dangularin
  real(kind=4) :: xmaxcone,xmincone

! Ninin treecode : Ntype=4
  real(kind=4) :: tnow,alpha_in,massp,aexp_in,omega0_in,omegaL_in,hubble_in
  real(kind=4) :: omega_lambda_t,omega_t
  real(kind=4), allocatable, dimension(:,:) :: tab

! MISC (junk formats)
  real(kind=4) :: myx,myy,myz,myt,myt2

  if (verbose) write(*,*) 'Read input particles file...'

  if ((Ntype == 2).and. &
      (outgtree.or.posttreatment)) then
     if (.not. simu_units) then
        write(*,*) 'physical units not yet possible for Ntype =',Ntype
        STOP
     endif
     if (posttreatment) then
        write(*,*) 'Action ='//posttreatm//' not yet possible for Ntype =',Ntype
        STOP
     endif
  endif

  if (Ntype==0) then
! PM unformatted real(kind=4)

     open(unit=lin,file=filein,form='unformatted',status='old')
     read(lin) alphalab
     read(lin) (alab(i),i=1,40)
     npart=nint(dble(alab(3))*dble(alab(4)))
     if (megaverbose) write(*,*) 'npart=',npart

     allocate(pos(3,npart))
     if (velocities) allocate(vel(3,npart))

     allocate(tmp(npart))
     if (megaverbose) write(*,*) 'read x...'
     read(lin) tmp
     pos(1,1:npart)=tmp(1:npart)
     if (megaverbose) write(*,*) 'read y...'
     read(lin) tmp
     pos(2,1:npart)=tmp(1:npart)
     if (megaverbose) write(*,*) 'read z...'
     read(lin) tmp
     pos(3,1:npart)=tmp(1:npart)
     if (megaverbose) write(*,*) 'read vx...'
     read(lin) tmp
     if (velocities) vel(1,1:npart)=tmp(1:npart)
     if (megaverbose) write(*,*) 'read vy...'
     read(lin) tmp
     if (velocities) vel(2,1:npart)=tmp(1:npart)
     if (megaverbose) write(*,*) 'read vz...'
     read(lin) tmp
     if (velocities) vel(3,1:npart)=tmp(1:npart)
     deallocate(tmp)
     read(lin) aexplu,itime,vmax,dt,wcin,wpot,rjns
     close(lin)

     omega0=dble(alab(26))
     omegaL=dble(alab(36))
     xlong =dble(alab(8))
     ylong =dble(alab(9))
     zlong =dble(alab(10))
     fnxlu =dble(alab(5))
     fnylu =dble(alab(6))
     fnzlu =dble(alab(7))
     amax  =dble(alab(32))
     hubble=dble(alab(30)) 

     aexp=dble(aexplu)/amax
     ! facteur multiplicatif pour convertir les positions en Mpc
     xmult =xlong/fnxlu

     ! facteur multiplicatif pour obtenir la vitesse propre physique
     vmult=1.5d0*amax**(-0.5d0)*(100.d0*hubble*xlong)/(fnxlu*aexp)

     ! Convert the positions in Mpc
     fnxfac=xlong/fnxlu
     fnyfac=ylong/fnylu
     fnzfac=zlong/fnzlu

     if (megaverbose) write(*,*) 'renormalize positions...'
     ! Positions are in Mpc, in a box of size xlong, ylong, zlong
     ! Origin of coordinates is at the center of the box
     pos_renorm=fnxfac
     xlongs2=0.5d0*xlong
     ylongs2=0.5d0*ylong
     zlongs2=0.5d0*zlong
     pos_shift(1)=-xlongs2
     pos_shift(2)=-ylongs2
     pos_shift(3)=-zlongs2
     pos(1,1:npart)=pos(1,1:npart)*fnxfac-xlongs2
     pos(2,1:npart)=pos(2,1:npart)*fnyfac-ylongs2
     pos(3,1:npart)=pos(3,1:npart)*fnzfac-zlongs2

     ! velocities in km/s
     if (velocities) then
        if (megaverbose) write(*,*) 'renormalize velocities'
        vel(1,1:npart)=vel(1,1:npart)*vmult
        vel(2,1:npart)=vel(2,1:npart)*vmult
        vel(3,1:npart)=vel(3,1:npart)*vmult
     endif
     velrenorm=vmult

     ! Compute mass of each particle 
     mass_in_kg=(xlong*ylong*zlong/dble(npart))*mega_parsec**3 &
 &             *omega0*critical_density*hubble**2
     allocate(masspart(npart))
     massrenorm=1.d0
     masspart(1:npart)=1.
     epsilon=fudgepsilon*fnxfac/dble(npart/(fnxlu*fnylu*fnzlu))**(1.d0/3.d0)

  elseif (Ntype==1) then
! Gadget

     npart=0
     do impi=0,abs(nmpi)-1
        if (nmpi==-1) then
           filempi=trim(filein) 
        else
           call convtoasc(impi,string)
           filempi=trim(filein)//string(6:6)
        endif
        if (megaverbose)  write(*,*) 'Read header of '//trim(filempi)
        open(unit=lin,file=filempi,form='unformatted',status='old')
        read(lin) (npart_in(i),i=0,5), (massarr_in(i),i=0,5), a_in,  &
 &             redshift_in, flag_sfr_in, flag_feedback_in,  &
 &             (nall_in(i),i=0,5), flag_cooling_in, numfiles_in, &
 &             xLbox_in,omega0,omegaL,hubble, &
 &             (unused_in(i),i=1,64-6-12-4-2*4-6*4-2*4-4*8)
        close(lin)
        npartp=sum(npart_in)
        npart=npart+npartp
        if (megaverbose) then          
           write(*,*) 'icpu,npartp,aexp=',impi,npartp,a_in
        endif
        if (numfiles_in > 1.and.abs(nmpi)==1) then
           write(*,*) 'ERROR in read_part :'
           write(*,*) 'Only pure dark matter simulations for GADGET'
           write(*,*) 'numfiles_in=',numfiles_in
           STOP
        endif
     enddo
        
     allocate(pos(3,npart))
     if (velocities) allocate(vel(3,npart))
     
     npart=0
     npartold=1
     do impi=0,abs(nmpi)-1
        if (nmpi==-1) then
           filempi=trim(filein) 
        else
           call convtoasc(impi,string)
           filempi=trim(filein)//string(6:6)
        endif
        open(unit=lin,file=filempi,form='unformatted',status='old')
        read(lin) (npart_in(i),i=0,5), (massarr_in(i),i=0,5), a_in,  &
 &             redshift_in, flag_sfr_in, flag_feedback_in,  &
 &             (nall_in(i),i=0,5), flag_cooling_in, numfiles_in, &
 &             xLbox_in,omega0,omegaL,hubble, &
 &             (unused_in(i),i=1,64-6-12-4-2*4-6*4-2*4-4*8)
        npart=npart+sum(npart_in)
        if (megaverbose)  write(*,*) 'Read '//trim(filempi)
        if (megaverbose) write(*,*) 'Read particle coordinates...'
        read(lin) ((pos(i,j),i=1,3),j=npartold,npart)
        if (velocities) then
           if (megaverbose) write(*,*) 'Read particle velocities...'
           read(lin) vel(1:3,npartold:npart)
        endif
        close(lin)     
        npartold=npart+1
     enddo

     aexp=a_in
     if (megaverbose) write(*,*) 'aexp=',aexp

     facco=1.d0/(1000.d0*hubble)
     xlong=xLbox_in*facco
     ylong=xlong
     zlong=xlong
     xlongs2=0.5d0*xlong
     ylongs2=0.5d0*ylong
     zlongs2=0.5d0*zlong
     ! Positions are in Mpc, in a box of size xlong, ylong, zlong
     ! Origin of coordinates is at the center of the box
     if (megaverbose) write(*,*) 'renormalize positions...'
     pos_renorm=facco
     pos_shift(1)=-xlongs2
     pos_shift(2)=-ylongs2
     pos_shift(3)=-zlongs2
     pos(1,1:npart)=pos(1,1:npart)*facco-xlongs2
     pos(2,1:npart)=pos(2,1:npart)*facco-ylongs2
     pos(3,1:npart)=pos(3,1:npart)*facco-zlongs2
     ! Convert velocities in physical peculiar velocities in km/s
     if (velocities) then
        if (megaverbose) write(*,*) 'renormalize velocities...'
        vel(1,1:npart)=vel(1,1:npart)*sqrt(a_in)
        vel(2,1:npart)=vel(2,1:npart)*sqrt(a_in)
        vel(3,1:npart)=vel(3,1:npart)*sqrt(a_in)
     endif
     velrenorm=sqrt(a_in)
     ! mass of each particle 
     mass_in_kg=(xlong*ylong*zlong/dble(npart))*mega_parsec**3 &
 &             *omega0*critical_density*hubble**2
     massrenorm=1.d0
     allocate(masspart(npart))
     masspart(1:npart)=1.
     ! fudgepsilon is of the order of the softening lenght in
     ! units of the mean interparticle distance
     epsilon=fudgepsilon*xlong/dble(npart)**(1.d0/3.d0)

  elseif (Ntype==2) then
! RAMSES unformatted cosmological multiple MPI files
     npart=0

     impi=1
     call convtoasc(impi,string)
     filempi=trim(filein)//string(2:6)
        
     open(unit=lin,file=filempi,form='unformatted',status='old')
     read(lin) ncpup
     close(lin)
     nmpi=ncpup
     
     do impi=1,nmpi
        call convtoasc(impi,string)
        filempi=trim(filein)//string(2:6)
        
        open(unit=lin,file=filempi,form='unformatted',status='old')
        read(lin) ncpup
        read(lin) ndimp
        read(lin) npartp
        if (megaverbose) write(*,*) 'icpu,ndimp,npartp=',impi,ndimp,npartp
        npart=npart+npartp
        close(lin)
     enddo
     if (ndimp /= 3) then
        write(*,*) 'ERROR in read_part : only 3D data possible.'
        STOP
     endif
     
     if (megaverbose)  write(*,*) 'Found ',npart,' particles'
     
     allocate(pos(3,npart))
     if (velocities) allocate(vel(3,npart))
     allocate(masspart(npart))
     if (mtot <= 0.0d0) then
        massrenorm=1.0d0
     else
        massrenorm=1.0d0/mtot
     endif
     if (Ltot <= 0.0d0) then
        xlong=1.0d0
     else
        xlong=Ltot
     endif
     ylong=xlong
     zlong=xlong
     xlongs2=0.5d0*xlong
     ylongs2=xlongs2
     zlongs2=xlongs2

     mymasstot=0.0d0
     npart=0
     do impi=1,nmpi
        call convtoasc(impi,string)
        filempi=trim(filein)//string(2:6)
        
        open(unit=lin,file=filempi,form='unformatted',status='old')
        read(lin) ncpup
        read(lin) ndimp
        read(lin) npartp
        if (megaverbose) write(*,*) 'Read file '//trim(filempi)
        allocate(tmpmpi(npartp))
        read(lin) tmpmpi
        pos(1,npart+1:npart+npartp)=tmpmpi(1:npartp)
        read(lin) tmpmpi
        pos(2,npart+1:npart+npartp)=tmpmpi(1:npartp)
        read(lin) tmpmpi
        pos(3,npart+1:npart+npartp)=tmpmpi(1:npartp)
        read(lin) tmpmpi
        if (velocities) then
           vel(1,npart+1:npart+npartp)=tmpmpi(1:npartp)
        endif
        read(lin) tmpmpi
        if (velocities) then
           vel(2,npart+1:npart+npartp)=tmpmpi(1:npartp)
        endif
        read(lin) tmpmpi
        if (velocities) then
           vel(3,npart+1:npart+npartp)=tmpmpi(1:npartp)
        endif
        read(lin) tmpmpi
        masspart(npart+1:npart+npartp)=tmpmpi(1:npartp)*massrenorm
        mymasstot=mymasstot+sum(dble(masspart(npart+1:npart+npartp)))
        deallocate(tmpmpi)
        close(lin)
        npart=npart+npartp
     enddo

     if (megaverbose) then
        write(*,*) 'x,y,z min=',minval(pos)
        write(*,*) 'x,y,z max=',maxval(pos)
        write(*,*) 'Total mass=',mymasstot
     endif
     
     npart_true=1

     pos_renorm=1.d0
     velrenorm=1.d0
     pos_shift(1)=-xlongs2
     pos_shift(2)=-ylongs2
     pos_shift(3)=-zlongs2
     do i=1,npart
        pos(1:3,i)=pos(1:3,i)+pos_shift(1:3)
     enddo
     epsilon=fudgepsilon*xlong/dble(npart)**(1.d0/3.d0)
     
  elseif (Ntype==3) then
! CONES 

     npart=nparmax
     allocate(pos(3,npart))
     if (velocities) allocate(vel(3,npart))
     allocate(masspart(npart))

     open(unit=lin,file=filein,form='formatted', status='old')
     read(lin,*) Nfiles
     inc=0
     do ifile=1,Nfiles
        read(lin,'(a)') filecone
        if (megaverbose) write(*,*) 'Reading file '//trim(filecone)        
        open(unit=lin2,file=filecone,status='old',form='unformatted')
        read(lin2) hubblein, omega0in, omegaLin
        read(lin2) zmaxconein, the1conein, the2conein, & 
 &                 thetashiftin, phishiftin, padin
        hubble=dble(hubblein)
        omega0=dble(omega0in)
        omegaL=dble(omegaLin)
        do 
           read(lin2,end=98) xin,yin,zin,vxin,vyin,vzin,redshiftin,dangularin
           inc=inc+1
           if (inc.gt.nparmax) then
              write(*,*) 'Nparmax is too small.'
              STOP
           endif
           if (inc==1) then
              xmaxcone=xin
              xmincone=xin
           else
              xmaxcone=max(xmaxcone,xin)
              xmincone=min(xmincone,xin)
           endif
           pos(1,inc)=xin
           pos(2,inc)=yin
           pos(3,inc)=zin
           if (velocities) then
              vel(1,inc)=vxin
              vel(2,inc)=vyin
              vel(3,inc)=vzin
           endif
           if (mod(inc,1000000)==0.and.megaverbose) write(*,*) xin,yin,zin, &
     &                 redshiftin
        enddo
98      continue
        close(lin2)
     enddo
     close(lin)
     npart=inc
     if (megaverbose) write(*,*) 'npart =',npart
     xlong=max(dble(xmaxcone)-dble(xmincone),boxsize)
     ylong=boxsize
     zlong=boxsize
     xlongs2=xlong*0.5d0
     ylongs2=ylong*0.5d0
     zlongs2=zlong*0.5d0
     pos_renorm=1.d0
     pos_shift(1)=-xmincone-xlongs2
     pos_shift(2)=0.d0
     pos_shift(3)=0.d0
     
     do i=1,npart
        pos(1,i)=pos(1,i)+pos_shift(1)
        pos(2,i)=pos(2,i)+pos_shift(2)
        pos(3,i)=pos(3,i)+pos_shift(3)
     enddo
     mass_in_kg=(dble(xmaxcone)-dble(xmincone))*ylong*zlong &
 &             /dble(npart)*mega_parsec**3*omega0*critical_density*hubble**2
     massrenorm=1.d0
     velrenorm=1.d0
     masspart(1:npart)=1.
     epsilon=fudgepsilon*((dble(xmaxcone)-dble(xmincone))*ylong*zlong &
 &          /dble(npart))**(1.d0/3.d0)
  elseif (Ntype==4) then
     ! Ninin treecode unformatted 
     if (velocities.or.posttreatment) then
        write(*,*) 'ERROR in read_part :'
        write(*,*) 'Velocities or posttreatment not yet allowed for Ntype=4'
        STOP
     endif
     open(unit=lin,file=filein,status='old',form='unformatted', &
 &                      access='sequential')

     read(lin) npart,tnow,alpha_in,massp,aexp_in,omega_t,omega0_in, &
 &             omegaL_in,omega_lambda_t,hubble_in
     aexp=aexp_in/aexp_max
     if (megaverbose) then
        write(*,*) 'npart,aexp=',npart,real(aexp)
        write(*,*) 'hubble,omega0,omegaL,boxsize2=', &
 &                 real(hubble),real(omega0),real(omegaL),real(boxsize2)
     endif
     xlong=boxsize2
     ylong=xlong
     zlong=xlong
     xlongs2=xlong*0.5d0
     ylongs2=ylong*0.5d0
     zlongs2=zlong*0.5d0
     pos_renorm=xlong
     pos_shift(1)=0.0d0
     pos_shift(2)=0.0d0
     pos_shift(3)=0.0d0
     allocate(tab(262144,3))
     allocate(pos(3,npart))
     do i=1,npart/262144
        read(lin) tab
        do j=1,262144
           pos(1,(i-1)*262144+j) = tab(j,1)*xlong
           pos(2,(i-1)*262144+j) = tab(j,2)*ylong
           pos(3,(i-1)*262144+j) = tab(j,3)*zlong
        enddo
     enddo
     write(*,*) 'xmin,xmax=',minval(pos(1,:)),maxval(pos(1,:))
     write(*,*) 'ymin,ymax=',minval(pos(2,:)),maxval(pos(2,:))
     write(*,*) 'zmin,zmax=',minval(pos(3,:)),maxval(pos(3,:))
     close(lin) 

     ! Compute mass of each particle
     mass_in_kg=(xlong*ylong*zlong/dble(npart))*mega_parsec**3 &
 &             *omega0*critical_density*hubble**2
     allocate(masspart(npart))
     massrenorm=1.d0
     masspart(1:npart)=1.
     epsilon=fudgepsilon*xlong/dble(npart)**(1.d0/3.d0)   
     deallocate(tab)
  elseif (Ntype==5) then
     xlong=1.0
     ylong=xlong
     zlong=xlong
     xlongs2=xlong*0.5d0
     ylongs2=ylong*0.5d0
     zlongs2=zlong*0.5d0
     pos_renorm=1.0
     pos_shift(1)=-xlongs2
     pos_shift(2)=-ylongs2
     pos_shift(3)=-zlongs2
     open(unit=lin,file=filein,status='old',form='unformatted')
     read(lin) npart
     if (megaverbose) write(*,*) 'npart=',npart
     allocate(pos(3,npart))
     allocate(tmp(npart))
     allocate(masspart(npart))
     read(lin) tmp
     pos(1,1:npart)=tmp(1:npart)*0.9999-0.5d0
     read(lin) tmp
     pos(2,1:npart)=tmp(1:npart)*0.9999-0.5d0
     read(lin) tmp
     pos(3,1:npart)=tmp(1:npart)*0.9999-0.5d0
     close(lin)
     deallocate(tmp)
     mass_in_kg=1.d0
     massrenorm=1.d0
     velrenorm=1.d0
     masspart(1:npart)=1.
     epsilon=fudgepsilon*xlong/dble(npart)**(1.d0/3.d0)
  elseif (Ntype==6 .or. Ntype==60) then
     if (velocities.or.posttreatment) then
        write(*,*) 'ERROR in read_part :'
        write(*,*) 'Velocities or posttreatment not yet allowed for Ntype=',Ntype
        STOP
     endif
     xlong=1.0
     ylong=xlong
     zlong=xlong
     xlongs2=xlong*0.5d0
     ylongs2=ylong*0.5d0
     zlongs2=zlong*0.5d0
     pos_renorm=1.0
     pos_shift(1)=-xlongs2
     pos_shift(2)=-ylongs2
     pos_shift(3)=-zlongs2
     open(unit=lin,file=filein,status='old',form='formatted')
     npart=0
     do
        if (Ntype==6) then
           read(lin,*,end=1000) myx,myy,myz,myt
           npart=npart+1
           if (npart <= 10 .and. megaverbose) write(*,*) myx,myy,myz,myt
        elseif (Ntype==60) then
           read(lin,*,end=1000) myx,myy,myz,myt,myt2
           if (myt==-1.0) npart=npart+1
           if (npart <= 10 .and. myt==-1.0 .and. megaverbose) write(*,*) myx,myy,myz,myt
        endif
     enddo
     
1000 continue

     if (verbose) write(*,*) 'Npart read=',npart
     allocate(pos(3,npart))
     allocate(masspart(npart))
     rewind(lin)
     npart=0
     do 
        if (Ntype==6) then
           read(lin,*,end=1001) myx,myy,myz,myt
           npart=npart+1
           pos(1,npart)=myx
           pos(2,npart)=myy
           pos(3,npart)=myz
        elseif (Ntype==60) then
           read(lin,*,end=1001) myx,myy,myz,myt,myt2
           if (myt==-1.0) then
              npart=npart+1
              pos(1,npart)=myx
              pos(2,npart)=myy
              pos(3,npart)=myz
           endif
        endif
     enddo
     
1001 continue
     close(lin)
     mass_in_kg=1.d0
     massrenorm=1.d0
     velrenorm=1.d0
     masspart(1:npart)=1.
     Npart_true=128**3
     epsilon=fudgepsilon*xlong/dble(Npart_true)**(1.d0/3.d0)     
  else
     write(*,*) 'ERROR in read_part : wrong value of Ntype.'
     STOP
  endif

  if (Ntype/=6 .and. Ntype /=2) Npart_true=Npart
end subroutine read_part

!=======================================================================
subroutine remove_degenerate_particles
!=======================================================================
  use compute_neiKDtree_common
  implicit none
  real(kind=4), parameter :: accurac=1.e-6

  integer, allocatable, dimension(:) :: idgene
  real(kind=4),allocatable, dimension(:) :: tmp
  real(kind=4),allocatable, dimension(:,:) :: possav
  logical,allocatable,dimension(:) :: move
  real(kind=4) :: xref,tolerance,tolerance2,phi,costeta,teta,sinteta
  real(kind=4) :: ran2
  logical :: doneiter,done
  integer :: i,niter,j,ipar,jpar,idd,jdd,incdege
  integer :: idum

  if (megaverbose) write(*,*) 'Move randomly particles at exactly the same position'

  idum=-111
  tolerance=max(xlongs2,ylongs2,zlongs2)*accurac
  allocate(idgene(npart),tmp(npart),move(npart))
  tolerance2=2.0*tolerance
  allocate(possav(3,npart))
  do i=1,npart
     possav(1:3,i)=pos(1:3,i)
  enddo

  niter=1
  doneiter=.false.
  do while (.not.doneiter) 
     if (megaverbose) write(*,*) 'Iteration :',niter
     niter=niter+1
     doneiter=.false.
     do i=1,npart
        tmp(i)=pos(1,i)
        move(i)=.false.
     enddo
     call indexx(npart,tmp,idgene)

     done=.false.
     i=1
     do while (.not.done)
        xref=tmp(idgene(i))
        j=i+1
        if (j > npart) then
           j=1
           done=.true.
        endif
        do while (abs(xref-tmp(idgene(j))) < tolerance)
           j=j+1
           if (j > npart) then
              j=1
              done=.true.
           endif
        enddo
        do ipar=i,j-1
           idd=idgene(ipar)
           do jpar=ipar+1,j-1
              jdd=idgene(jpar)
              if (abs(pos(2,jdd)-pos(2,idd)) < tolerance .and. &
 &                abs(pos(3,jdd)-pos(3,idd)) < tolerance) then
                 move(idgene(jpar))=.true.
              endif
           enddo
        enddo
        i=j
     enddo
     incdege=0
     do i=1,npart
        if (move(i)) then
           incdege=incdege+1
        endif
     enddo
     if (megaverbose) write(*,*) &
 &                 'Found the following number of degeneracies :',incdege
     do i=1,npart
        if (move(i)) then
           phi=2.*pi*ran2(idum)
           costeta=1.0-2.0*ran2(idum)
           teta=acos(costeta)           
           sinteta=sin(teta)
           pos(1,i)=possav(1,i)+tolerance2*cos(phi)*sinteta
           pos(2,i)=possav(2,i)+tolerance2*sin(phi)*sinteta
           pos(3,i)=possav(3,i)+tolerance2*costeta
        endif
     enddo
     
     if (incdege==0) doneiter=.true.
  enddo

  deallocate(possav)
  deallocate(tmp,idgene,move)
  
end subroutine remove_degenerate_particles

!=======================================================================
subroutine convtoasc(number,sstring)
!=======================================================================
! To convert an integer smaller than 999999 to a 6 characters string
!=======================================================================
  implicit none
  integer :: number, istring, num, nums10, i
  character*6 :: sstring
  character*10,parameter :: nstring='0123456789'
      
  num=1000000
  nums10=num/10
  do i=1,6
     istring=1+mod(number,num)/nums10
     sstring(i:i)=nstring(istring:istring)
     num=num/10
     nums10=nums10/10
  enddo
end subroutine convtoasc

!=======================================================================
function HsurH0(z,omega0,omegaL,omegaR)
!=======================================================================
  implicit none
  real(kind=8) :: z,omega0,omegaL,omegaR,HsurH0

  HsurH0=sqrt(Omega0*(1.d0+z)**3+OmegaR*(1.d0+z)**2+OmegaL)
end function HsurH0

!=======================================================================
subroutine indexx(n,arr,indx)
!=======================================================================
! From Numerical recipes
!=======================================================================
  integer :: n,indx(n),M,NSTACK
  real(kind=4) :: arr(n)
  parameter (M=7,NSTACK=50)
  integer :: i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
  real(kind=4) :: a
  do j=1,n
     indx(j)=j
  enddo
  jstack=0
  l=1
  ir=n
1 if(ir-l.lt.M)then
     do j=l+1,ir
        indxt=indx(j)
        a=arr(indxt)
        do i=j-1,1,-1
           if(arr(indx(i)).le.a)goto 2
           indx(i+1)=indx(i)
        enddo
        i=0
2       indx(i+1)=indxt
     enddo
     if(jstack==0)return
     ir=istack(jstack)
     l=istack(jstack-1)
     jstack=jstack-2
  else
     k=(l+ir)/2
     itemp=indx(k)
     indx(k)=indx(l+1)
     indx(l+1)=itemp
     if(arr(indx(l+1)).gt.arr(indx(ir)))then
        itemp=indx(l+1)
        indx(l+1)=indx(ir)
        indx(ir)=itemp
     endif
     if(arr(indx(l)).gt.arr(indx(ir)))then
        itemp=indx(l)
        indx(l)=indx(ir)
        indx(ir)=itemp
     endif
     if(arr(indx(l+1)).gt.arr(indx(l)))then
        itemp=indx(l+1)
        indx(l+1)=indx(l)
        indx(l)=itemp
     endif
     i=l+1
     j=ir
     indxt=indx(l)
     a=arr(indxt)
3    continue
     i=i+1
     if(arr(indx(i)).lt.a)goto 3
4    continue
     j=j-1
     if(arr(indx(j)).gt.a)goto 4
     if(j.lt.i)goto 5
     itemp=indx(i)
     indx(i)=indx(j)
     indx(j)=itemp
     goto 3
5    indx(l)=indx(j)
     indx(j)=indxt
     jstack=jstack+2
     if(jstack.gt.NSTACK)pause 'NSTACK too small in indexx'
     if(ir-i+1.ge.j-l)then
        istack(jstack)=ir
        istack(jstack-1)=i
        ir=j-1
     else
        istack(jstack)=j-1
        istack(jstack-1)=l
        l=i
     endif
  endif
  goto 1
end subroutine indexx

!=======================================================================
subroutine int_indexx(n,arr,indx)
!=======================================================================
! From Numerical recipes
!=======================================================================
  integer :: n,indx(n),M,NSTACK
  integer :: arr(n)
  parameter (M=7,NSTACK=50)
  integer :: i,indxt,ir,itemp,j,jstack,k,l,istack(NSTACK)
  integer :: a
  do j=1,n
     indx(j)=j
  enddo
  jstack=0
  l=1
  ir=n
1 if(ir-l.lt.M)then
     do j=l+1,ir
        indxt=indx(j)
        a=arr(indxt)
        do i=j-1,1,-1
           if(arr(indx(i)).le.a)goto 2
           indx(i+1)=indx(i)
        enddo
        i=0
2       indx(i+1)=indxt
     enddo
     if(jstack==0)return
     ir=istack(jstack)
     l=istack(jstack-1)
     jstack=jstack-2
  else
     k=(l+ir)/2
     itemp=indx(k)
     indx(k)=indx(l+1)
     indx(l+1)=itemp
     if(arr(indx(l+1)).gt.arr(indx(ir)))then
        itemp=indx(l+1)
        indx(l+1)=indx(ir)
        indx(ir)=itemp
     endif
     if(arr(indx(l)).gt.arr(indx(ir)))then
        itemp=indx(l)
        indx(l)=indx(ir)
        indx(ir)=itemp
     endif
     if(arr(indx(l+1)).gt.arr(indx(l)))then
        itemp=indx(l+1)
        indx(l+1)=indx(l)
        indx(l)=itemp
     endif
     i=l+1
     j=ir
     indxt=indx(l)
     a=arr(indxt)
3    continue
     i=i+1
     if(arr(indx(i)).lt.a)goto 3
4    continue
     j=j-1
     if(arr(indx(j)).gt.a)goto 4
     if(j.lt.i)goto 5
     itemp=indx(i)
     indx(i)=indx(j)
     indx(j)=itemp
     goto 3
5    indx(l)=indx(j)
     indx(j)=indxt
     jstack=jstack+2
     if(jstack.gt.NSTACK)pause 'NSTACK too small in indexx'
     if(ir-i+1.ge.j-l)then
        istack(jstack)=ir
        istack(jstack-1)=i
        ir=j-1
     else
        istack(jstack)=j-1
        istack(jstack-1)=l
        l=i
     endif
  endif
  goto 1
end subroutine int_indexx

!=======================================================================
FUNCTION ran2(idum)
!=======================================================================
! From Numerical recipes
!=======================================================================
  integer :: idum,IM1,IM2,IMM1,IA1,IA2,IQ1,IQ2,IR1,IR2,NTAB,NDIV
  real(kind=4) :: ran2,AM,EPS,RNMX
  parameter (IM1=2147483563,IM2=2147483399,AM=1./IM1,IMM1=IM1-1, &
 &   IA1=40014,IA2=40692,IQ1=53668,IQ2=52774,IR1=12211,IR2=3791, &
 &   NTAB=32,NDIV=1+IMM1/NTAB,EPS=1.2e-7,RNMX=1.-EPS)
  integer :: idum2,j,k,iv(NTAB),iy
  save iv,iy,idum2
  data idum2/123456789/, iv/NTAB*0/, iy/0/

  if (idum.le.0) then
     idum=max(-idum,1)
     idum2=idum
     do j=NTAB+8,1,-1
        k=idum/IQ1
        idum=IA1*(idum-k*IQ1)-k*IR1
        if (idum.lt.0) idum=idum+IM1
        if (j.le.NTAB) iv(j)=idum
     enddo
     iy=iv(1)
  endif
  k=idum/IQ1
  idum=IA1*(idum-k*IQ1)-k*IR1
  if (idum.lt.0) idum=idum+IM1
  k=idum2/IQ2
  idum2=IA2*(idum2-k*IQ2)-k*IR2
  if (idum2.lt.0) idum2=idum2+IM2
  j=1+iy/NDIV
  iy=iv(j)-idum2
  iv(j)=idum
  if(iy.lt.1)iy=iy+IMM1
  ran2=min(AM*iy,RNMX)
end FUNCTION ran2
